"""Test the computation of form signatures."""

from ufl import (
    Argument,
    CellDiameter,
    CellVolume,
    Circumradius,
    Coefficient,
    FacetArea,
    FacetNormal,
    FunctionSpace,
    Identity,
    Mesh,
    SpatialCoordinate,
    TestFunction,
    as_vector,
    diff,
    dot,
    ds,
    dx,
    hexahedron,
    indices,
    inner,
    interval,
    quadrilateral,
    tetrahedron,
    triangle,
    variable,
)
from ufl.algorithms.signature import compute_multiindex_hashdata, compute_terminal_hashdata
from ufl.classes import FixedIndex, MultiIndex
from ufl.finiteelement import FiniteElement, SymmetricElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2

# TODO: Test compute_terminal_hashdata
#   TODO: Check that form argument counts only affect the sig by their relative ordering
#   TODO: Check that all other relevant terminal propeties affect the terminal_hashdata

# TODO: Test that operator types affect the sig
# TODO: Test that we do not get collisions for some large sets of generated forms
# TODO: How do we know that we have tested the signature reliably enough?


def domain_numbering(*cells):
    renumbering = {}
    for i, cell in enumerate(cells):
        d = cell.topological_dimension()
        domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1), ufl_id=i)
        renumbering[domain] = i
    return renumbering


def compute_unique_terminal_hashdatas(hashdatas):
    count = 0
    data = set()
    hashes = set()
    reprs = set()
    for d in hashdatas:
        # Each d is the result of a compute_terminal_hashdatas call,
        # which is a dict where the keys are non-canonical terminals
        # and the values are the canonical hashdata.
        # We want to count unique hashdata values,
        # ignoring the original terminals.
        assert isinstance(d, dict)
        # Sorting values by hash should be stable at least in a single test run:
        t = tuple(sorted(list(d.values()), key=lambda x: hash(x)))
        # print t

        # Add the hashdata values tuple to sets based on itself, its hash,
        # and its repr (not sure why I included repr anymore?)
        hashes.add(hash(t))  # This will fail if t is not hashable, which it should be!
        data.add(t)
        reprs.add(repr(t))
        count += 1

    return count, len(data), len(reprs), len(hashes)


def test_terminal_hashdata_depends_on_literals(self):
    reprs = set()
    hashes = set()

    def forms():
        i, j = indices(2)
        for d, cell in [(2, triangle), (3, tetrahedron)]:
            domain = Mesh(
                FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1), ufl_id=d - 2
            )
            x = SpatialCoordinate(domain)
            ident = Identity(d)
            for fv in (1.1, 2.2):
                for iv in (5, 7):
                    expr = (ident[0, j] * (fv * x[j])) ** iv

                    reprs.add(repr(expr))
                    hashes.add(hash(expr))
                    yield compute_terminal_hashdata(expr, {domain: 0})

    c, d, r, h = compute_unique_terminal_hashdatas(forms())
    assert c == 8
    assert d == c
    assert r == c
    assert h == c
    assert len(reprs) == c
    assert len(hashes) == c


def test_terminal_hashdata_depends_on_geometry(self):
    reprs = set()
    hashes = set()

    def forms():
        i, j = indices(2)
        cells = (triangle, tetrahedron)
        for i, cell in enumerate(cells):
            d = cell.topological_dimension()
            domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1), ufl_id=i)

            x = SpatialCoordinate(domain)
            n = FacetNormal(domain)
            h = CellDiameter(domain)
            r = Circumradius(domain)
            a = FacetArea(domain)
            # s = CellSurfaceArea(domain)
            v = CellVolume(domain)
            ident = Identity(d)

            ws = (x, n)
            qs = (h, r, a, v)  # , s)
            for w in ws:
                for q in qs:
                    expr = ident[0, j] * (q * w[j])

                    reprs.add(repr(expr))
                    hashes.add(hash(expr))
                    yield compute_terminal_hashdata(expr, domain_numbering(*cells))

    c, d, r, h = compute_unique_terminal_hashdatas(forms())
    assert c == 2 * 4 * 2  # len(ws)*len(qs)*len(cells)
    assert d == c
    assert r == c
    assert h == c
    assert len(reprs) == c
    assert len(hashes) == c


def test_terminal_hashdata_depends_on_form_argument_properties(self):
    reprs = set()
    hashes = set()
    nelm = 5
    nreps = 2

    # Data
    cells = (triangle, tetrahedron)
    degrees = (1, 2)
    families = (("Lagrange", H1), ("Lagrange", H1), ("Discontinuous Lagrange", L2))

    def forms():
        for rep in range(nreps):
            for i, cell in enumerate(cells):
                d = cell.topological_dimension()
                domain = Mesh(
                    FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1), ufl_id=i
                )
                for degree in degrees:
                    for family, sobolev in families:
                        V = FiniteElement(family, cell, degree, (), identity_pullback, sobolev)
                        W = FiniteElement(family, cell, degree, (d,), identity_pullback, sobolev)
                        W2 = FiniteElement(
                            family, cell, degree, (d + 1,), identity_pullback, sobolev
                        )
                        T = FiniteElement(family, cell, degree, (d, d), identity_pullback, sobolev)
                        if d == 2:
                            S = SymmetricElement(
                                {(0, 0): 0, (0, 1): 1, (1, 0): 1, (1, 1): 2},
                                [
                                    FiniteElement(
                                        family, cell, degree, (), identity_pullback, sobolev
                                    )
                                    for _ in range(3)
                                ],
                            )
                        else:
                            assert d == 3
                            S = SymmetricElement(
                                {
                                    (0, 0): 0,
                                    (0, 1): 1,
                                    (0, 2): 2,
                                    (1, 0): 1,
                                    (1, 1): 3,
                                    (1, 2): 4,
                                    (2, 0): 2,
                                    (2, 1): 4,
                                    (2, 2): 5,
                                },
                                [
                                    FiniteElement(
                                        family, cell, degree, (), identity_pullback, sobolev
                                    )
                                    for _ in range(6)
                                ],
                            )
                        elements = [V, W, W2, T, S]
                        assert len(elements) == nelm

                        for H in elements[:nelm]:
                            space = FunctionSpace(domain, H)
                            # Keep number and count fixed, we're not testing that here
                            a = Argument(space, number=1)
                            c = Coefficient(space, count=1)
                            renumbering = domain_numbering(*cells)
                            renumbering[c] = 0
                            for f in (a, c):
                                expr = inner(f, f)

                                reprs.add(repr(expr))
                                hashes.add(hash(expr))
                                yield compute_terminal_hashdata(expr, renumbering)

    c, d, r, h = compute_unique_terminal_hashdatas(forms())
    c1 = nreps * len(cells) * len(degrees) * len(families) * nelm * 2
    assert c == c1

    c0 = len(cells) * len(degrees) * (len(families) - 1) * nelm * 2
    assert d == c0
    assert r == c0
    assert h == c0
    assert len(reprs) == c0
    assert len(hashes) == c0


def test_terminal_hashdata_does_not_depend_on_coefficient_count_values_only_ordering(self):
    reprs = set()
    hashes = set()
    counts = list(range(-3, 4))
    cells = (interval, triangle, hexahedron)
    assert len(counts) == 7
    nreps = 1

    def forms():
        for rep in range(nreps):
            for i, cell in enumerate(cells):
                d = cell.topological_dimension()
                domain = Mesh(
                    FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1), ufl_id=i
                )
                for k in counts:
                    V = FiniteElement("Lagrange", cell, 2, (), identity_pullback, H1)
                    space = FunctionSpace(domain, V)
                    f = Coefficient(space, count=k)
                    g = Coefficient(space, count=k + 2)
                    expr = inner(f, g)

                    renumbering = domain_numbering(*cells)
                    renumbering[f] = 0
                    renumbering[g] = 1

                    reprs.add(repr(expr))
                    hashes.add(hash(expr))
                    yield compute_terminal_hashdata(expr, renumbering)

    c, d, r, h = compute_unique_terminal_hashdatas(forms())
    c0 = len(cells)  # Number of actually unique cases from a code generation perspective
    c1 = len(counts) * c0  # Number of unique cases from a symbolic representation perspective
    assert len(reprs) == c1
    assert len(hashes) == c1
    assert c == nreps * c1  # number of inner loop executions in forms() above
    assert d == c0
    assert r == c0
    assert h == c0


def test_terminal_hashdata_does_depend_on_argument_number_values(self):
    # TODO: Include part numbers as well
    reprs = set()
    hashes = set()
    counts = list(range(4))
    cells = (interval, triangle, hexahedron)
    nreps = 2

    def forms():
        for rep in range(nreps):
            for i, cell in enumerate(cells):
                d = cell.topological_dimension()
                domain = Mesh(
                    FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1), ufl_id=i
                )
                for k in counts:
                    V = FiniteElement("Lagrange", cell, 2, (), identity_pullback, H1)
                    space = FunctionSpace(domain, V)
                    f = Argument(space, k)
                    g = Argument(space, k + 2)
                    expr = inner(f, g)

                    reprs.add(repr(expr))
                    hashes.add(hash(expr))
                    yield compute_terminal_hashdata(expr, domain_numbering(*cells))

    c, d, r, h = compute_unique_terminal_hashdatas(forms())
    c0 = len(cells) * len(
        counts
    )  # Number of actually unique cases from a code generation perspective
    c1 = 1 * c0  # Number of unique cases from a symbolic representation perspective
    assert len(reprs) == c1
    assert len(hashes) == c1
    self.assertEqual(c, nreps * c1)  # number of inner loop executions in forms() above
    assert d == c0
    assert r == c0
    assert h == c0


def test_domain_signature_data_does_not_depend_on_domain_label_value(self):
    cells = [triangle, tetrahedron, hexahedron]
    s0s = set()
    s1s = set()
    s2s = set()
    for i, cell in enumerate(cells):
        d = cell.topological_dimension()
        domain = FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1)
        d0 = Mesh(domain)
        d1 = Mesh(domain, ufl_id=1)
        d2 = Mesh(domain, ufl_id=2)
        s0 = d0._ufl_signature_data_({d0: 0})
        s1 = d1._ufl_signature_data_({d1: 0})
        s2 = d2._ufl_signature_data_({d2: 0})
        assert s0 == s1
        assert s0 == s2
        s0s.add(s0)
        s1s.add(s1)
        s2s.add(s2)
    assert len(s0s) == len(cells)
    assert len(s1s) == len(cells)
    assert len(s2s) == len(cells)


def test_terminal_hashdata_does_not_depend_on_domain_label_value(self):
    reprs = set()
    hashes = set()
    ufl_ids = [1, 2]
    cells = [triangle, quadrilateral]
    domains = [
        Mesh(
            FiniteElement(
                "Lagrange", cell, 1, (cell.topological_dimension(),), identity_pullback, H1
            ),
            ufl_id=ufl_id,
        )
        for cell in cells
        for ufl_id in ufl_ids
    ]
    nreps = 2
    num_exprs = 2

    def forms():
        for rep in range(nreps):
            for domain in domains:
                V = FunctionSpace(
                    domain,
                    FiniteElement("Lagrange", domain.ufl_cell(), 2, (), identity_pullback, H1),
                )
                f = Coefficient(V, count=0)
                v = TestFunction(V)
                x = SpatialCoordinate(domain)
                n = FacetNormal(domain)
                exprs = [inner(x, n), inner(f, v)]
                assert num_exprs == len(exprs)  # Assumed in checks below

                # This numbering needs to be recreated to count 'domain' and 'f' as 0 each time:
                renumbering = {f: 0, domain: 0}

                for expr in exprs:
                    reprs.add(repr(expr))
                    hashes.add(hash(expr))
                    yield compute_terminal_hashdata(expr, renumbering)

    c, d, r, h = compute_unique_terminal_hashdatas(forms())
    c0 = num_exprs * len(
        cells
    )  # Number of actually unique cases from a code generation perspective
    c1 = num_exprs * len(
        domains
    )  # Number of unique cases from a symbolic representation perspective
    assert len(reprs) == c1
    assert len(hashes) == c1
    self.assertEqual(c, nreps * c1)  # number of inner loop executions in forms() above
    assert d == c0
    assert r == c0
    assert h == c0


def compute_unique_multiindex_hashdatas(hashdatas):
    count = 0
    data = set()
    hashes = set()
    reprs = set()
    for d in hashdatas:
        data.add(tuple(d))
        hashes.add(hash(tuple(d)))
        reprs.add(repr(d))
        count += 1
    return count, len(data), len(reprs), len(hashes)


def test_multiindex_hashdata_depends_on_fixed_index_values(self):
    reprs = set()
    hashes = set()

    def hashdatas():
        for i in range(3):
            for ii in ((i,), (i, 0), (1, i)):
                jj = tuple(FixedIndex(j) for j in ii)
                expr = MultiIndex(jj)
                reprs.add(repr(expr))
                hashes.add(hash(expr))
                yield compute_multiindex_hashdata(expr, {})

    c, d, r, h = compute_unique_multiindex_hashdatas(hashdatas())
    assert c == 9
    assert d == 9 - 1  # (1,0 is repeated, therefore -1)
    assert len(reprs) == 9 - 1
    assert len(hashes) == 9 - 1


def test_multiindex_hashdata_does_not_depend_on_counts(self):
    reprs = set()
    hashes = set()

    def hashdatas():
        ijs = []
        iind = indices(3)
        jind = indices(3)
        for i in iind:
            ijs.append((i,))
            for j in jind:
                ijs.append((i, j))
                ijs.append((j, i))
        for ij in ijs:
            expr = MultiIndex(ij)
            reprs.add(repr(expr))
            hashes.add(hash(expr))
            yield compute_multiindex_hashdata(expr, {})

    c, d, r, h = compute_unique_multiindex_hashdatas(hashdatas())
    assert c == 3 + 9 + 9
    assert d == 1 + 1
    assert len(reprs) == 3 + 9 + 9
    assert len(hashes) == 3 + 9 + 9


def test_multiindex_hashdata_depends_on_the_order_indices_are_observed(self):
    reprs = set()
    hashes = set()
    nrep = 3

    def hashdatas():
        for rep in range(nrep):
            # Resetting index_numbering for each repetition, resulting
            # in hashdata staying the same for each repetition but repr
            # and hashes changing because new indices are created each
            # repetition.
            index_numbering = {}
            i, j, k, l = indices(4)  # noqa: E741
            for expr in (
                MultiIndex((i,)),
                MultiIndex((i,)),  # r
                MultiIndex((i, j)),
                MultiIndex((j, i)),
                MultiIndex((i, j)),  # r
                MultiIndex((i, j, k)),
                MultiIndex((k, j, i)),
                MultiIndex((j, i)),
            ):  # r
                reprs.add(repr(expr))
                hashes.add(hash(expr))
                yield compute_multiindex_hashdata(expr, index_numbering)

    c, d, r, h = compute_unique_multiindex_hashdatas(hashdatas())
    assert c == nrep * 8
    assert d == 5
    assert len(reprs) == nrep * 5
    assert len(hashes) == nrep * 5


def check_unique_signatures(forms):
    count = 0
    sigs = set()
    sigs2 = set()
    hashes = set()
    reprs = set()
    for a in forms:
        sig = a.signature()
        sig2 = a.signature()
        sigs.add(sig)
        sigs2.add(sig2)
        assert sig
        hashes.add(hash(a))
        reprs.add(repr(a))
        count += 1
    assert len(sigs) == count
    assert len(sigs2) == count
    assert len(reprs) == count
    assert len(hashes) == count


def test_signature_is_affected_by_element_properties(self):
    def forms():
        for family, sobolev in (("Lagrange", H1), ("Discontinuous Lagrange", L2)):
            for cell in (triangle, tetrahedron, quadrilateral):
                d = cell.topological_dimension()
                domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))
                for degree in (1, 2):
                    V = FiniteElement(family, cell, degree, (), identity_pullback, sobolev)
                    space = FunctionSpace(domain, V)
                    u = Coefficient(space)
                    v = TestFunction(space)
                    x = SpatialCoordinate(domain)
                    w = as_vector([v] * x.ufl_shape[0])
                    f = dot(w, u * x)
                    a = f * dx
                    yield a

    check_unique_signatures(forms())


def test_signature_is_affected_by_domains(self):
    def forms():
        for cell in (triangle, tetrahedron):
            d = cell.topological_dimension()
            domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))
            for di in (1, 2):
                for dj in (1, 2):
                    for dk in (1, 2):
                        V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
                        space = FunctionSpace(domain, V)
                        u = Coefficient(space)
                        a = u * dx(di) + 2 * u * dx(dj) + 3 * u * ds(dk)
                        yield a

    check_unique_signatures(forms())


def test_signature_of_forms_with_diff(self):
    def forms():
        for i, cell in enumerate([triangle, tetrahedron]):
            d = cell.topological_dimension()
            domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1), ufl_id=i)
            for k in (1, 2, 3):
                d = cell.topological_dimension()
                V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
                W = FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1)
                v_space = FunctionSpace(domain, V)
                w_space = FunctionSpace(domain, W)
                u = Coefficient(v_space)
                w = Coefficient(w_space)
                vu = variable(u)
                vw = variable(w)
                f = vu * dot(vw, vu**k * vw)
                g = diff(f, vu)
                h = dot(diff(f, vw), FacetNormal(domain))
                a = f * dx(1) + g * dx(2) + h * ds(0)
                yield a

    check_unique_signatures(forms())


def test_signature_of_form_depend_on_coefficient_numbering_across_integrals(self):
    cell = triangle
    V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    f = Coefficient(space)
    g = Coefficient(space)
    M1 = f * dx(0) + g * dx(1)
    M2 = g * dx(0) + f * dx(1)
    M3 = g * dx(0) + g * dx(1)
    self.assertTrue(M1.signature() != M2.signature())
    self.assertTrue(M1.signature() != M3.signature())
    self.assertTrue(M2.signature() != M3.signature())


def test_signature_of_forms_change_with_operators(self):
    def forms():
        for cell in (triangle, tetrahedron):
            d = cell.topological_dimension()
            V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
            domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))
            space = FunctionSpace(domain, V)
            u = Coefficient(space)
            v = Coefficient(space)
            fs = [
                (u * v) + (u / v),
                (u + v) + (u / v),
                (u + v) * (u / v),
                (u * v) * (u * v),
                (u + v) * (u * v),  # H1 same
                # (u*v)*(u+v), # H1 same
                (u * v) + (u + v),
            ]
            for f in fs:
                a = f * dx
                yield a

    check_unique_signatures(forms())



import pytest

from ufl import (
    Coefficient,
    FacetNormal,
    FunctionSpace,
    Mesh,
    SpatialCoordinate,
    TestFunction,
    TrialFunction,
    adjoint,
    cofac,
    conj,
    derivative,
    ds,
    dx,
    grad,
    inner,
    tetrahedron,
)
from ufl.algorithms.check_arities import ArityMismatch
from ufl.algorithms.compute_form_data import compute_form_data
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_check_arities():
    # Code from bitbucket issue #49
    cell = tetrahedron
    D = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
    V = FunctionSpace(D, FiniteElement("Lagrange", cell, 2, (3,), identity_pullback, H1))
    dv = TestFunction(V)
    du = TrialFunction(V)

    X = SpatialCoordinate(D)
    N = FacetNormal(D)

    u = Coefficient(V)
    x = X + u
    F = grad(x)
    n = cofac(F) * N

    M = inner(x, n) * ds
    L = derivative(M, u, dv)
    a = derivative(L, u, du)

    compute_form_data(M)
    compute_form_data(L)
    compute_form_data(a)


def test_complex_arities():
    cell = tetrahedron
    D = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
    V = FunctionSpace(D, FiniteElement("Lagrange", cell, 2, (3,), identity_pullback, H1))
    v = TestFunction(V)
    u = TrialFunction(V)

    # Valid form.
    F = inner(u, v) * dx
    compute_form_data(F, complex_mode=True)
    # Check that adjoint conjugates correctly
    compute_form_data(adjoint(F), complex_mode=True)

    with pytest.raises(ArityMismatch):
        compute_form_data(inner(v, u) * dx, complex_mode=True)

    with pytest.raises(ArityMismatch):
        compute_form_data(inner(conj(v), u) * dx, complex_mode=True)


def test_product_arity():
    cell = tetrahedron
    D = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
    V = FunctionSpace(D, FiniteElement("Lagrange", cell, 2, (3,), identity_pullback, H1))
    v = TestFunction(V)
    u = TrialFunction(V)

    with pytest.raises(ArityMismatch):
        F = inner(u, u) * dx
        compute_form_data(F, complex_mode=True)

    with pytest.raises(ArityMismatch):
        L = inner(v, v) * dx
        compute_form_data(L, complex_mode=False)



__authors__ = "Martin Sandve Alnæs"
__date__ = "2008-08-20 -- 2012-11-30"

import pytest

from ufl import Coefficient, FunctionSpace, Mesh, conditional, eq, ge, gt, le, lt, ne, triangle
from ufl.classes import EQ, GE, GT, LE, LT, NE
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


@pytest.fixture
def f():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    return Coefficient(space)


@pytest.fixture
def g():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    return Coefficient(space)


def test_conditional_does_not_allow_bool_condition(f, g):
    # The reason for this test is that it protects from the case
    # conditional(a == b, t, f) in which a == b means comparing representations
    with pytest.raises(BaseException):
        conditional(True, 1, 0)


def test_eq_oper_produces_bool(f, g):
    expr1 = f == f
    expr2 = f == g
    assert isinstance(expr1, bool)
    assert isinstance(expr2, bool)
    assert expr1
    assert not expr2


def test_eq_produces_ufl_expr(f, g):
    expr1 = eq(f, g)
    expr2 = eq(f, f)
    expr3 = f == g
    expr4 = f == f
    # Correct types:
    assert isinstance(expr1, EQ)
    assert isinstance(expr2, EQ)
    assert isinstance(expr3, bool)
    assert isinstance(expr4, bool)
    # Comparing representations correctly:
    assert bool(expr1 == eq(f, g))
    assert bool(expr1 != eq(g, g))
    assert bool(expr2 == eq(f, f))
    assert bool(expr2 != eq(g, f))
    # Bool evaluation yields actual bools:
    assert isinstance(bool(expr1), bool)
    assert isinstance(bool(expr2), bool)
    assert not expr3
    assert expr4
    # Allow use in boolean python expression context:
    # NB! This means comparing representations! Required by dict and set.
    assert not bool(expr1)
    assert bool(expr2)
    assert not bool(expr3)
    assert bool(expr4)


def test_ne_produces_ufl_expr(f, g):
    expr1 = ne(f, g)
    expr2 = ne(f, f)
    expr3 = f != g
    expr4 = f != f
    # Correct types:
    assert isinstance(expr1, NE)
    assert isinstance(expr2, NE)
    assert isinstance(expr3, bool)
    assert isinstance(expr4, bool)
    # Comparing representations correctly:
    assert bool(expr1 == ne(f, g))
    assert bool(expr1 != ne(g, g))
    assert bool(expr2 == ne(f, f))
    assert bool(expr2 != ne(g, f))
    assert not bool(expr2 == expr3)
    # Bool evaluation yields actual bools:
    assert isinstance(bool(expr1), bool)
    assert isinstance(bool(expr2), bool)
    # Allow use in boolean python expression context:
    # NB! This means the opposite of ==, i.e. comparing representations!
    assert bool(expr1)
    assert not bool(expr2)
    assert bool(expr1)
    assert not bool(expr2)


def test_lt_produces_ufl_expr(f, g):
    expr1 = lt(f, g)
    expr2 = f < g
    # Correct types (no bools here!):
    assert isinstance(expr1, LT)
    assert isinstance(expr2, LT)
    # Representations are the same:
    assert bool(expr1 == expr2)
    # Protection from misuse in boolean python expression context:
    with pytest.raises(BaseException):
        bool(expr1)


def test_gt_produces_ufl_expr(f, g):
    expr1 = gt(f, g)
    expr2 = f > g
    # Correct types (no bools here!):
    assert isinstance(expr1, GT)
    assert isinstance(expr2, GT)
    # Representations are the same:
    assert bool(expr1 == expr2)
    # Protection from misuse in boolean python expression context:
    with pytest.raises(BaseException):
        bool(expr1)


def test_le_produces_ufl_expr(f, g):
    expr1 = le(f, g)
    expr2 = f <= g
    # Correct types (no bools here!):
    assert isinstance(expr1, LE)
    assert isinstance(expr2, LE)
    # Representations are the same:
    assert bool(expr1 == expr2)
    # Protection from misuse in boolean python expression context:
    with pytest.raises(BaseException):
        bool(expr1)


def test_ge_produces_ufl_expr(f, g):
    expr1 = ge(f, g)
    expr2 = f >= g
    # Correct types (no bools here!):
    assert isinstance(expr1, GE)
    assert isinstance(expr2, GE)
    # Representations are the same:
    assert bool(expr1 == expr2)
    # Protection from misuse in boolean python expression context:
    with pytest.raises(BaseException):
        bool(expr1)



"""This module contains algorithms for traversing expression trees in different ways."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008

from ufl.action import Action
from ufl.adjoint import Adjoint
from ufl.core.expr import Expr
from ufl.form import BaseForm, Form, FormSum
from ufl.integral import Integral


def iter_expressions(a):
    """Handle Form, Integral and any Expr the same way when inspecting expressions.

    Returns an iterable over Expr instances:
    - a is an Expr: (a,)
    - a is an Integral:  the integrand expression of a
    - a is a  Form:      all integrand expressions of all integrals
    - a is a  FormSum:   the components of a
    - a is an Action:    the left and right component of a
    - a is an Adjoint:   the underlying form of a
    """
    if isinstance(a, Form):
        return (itg.integrand() for itg in a.integrals())
    elif isinstance(a, Integral):
        return (a.integrand(),)
    elif isinstance(a, (FormSum, Adjoint, Action)):
        return tuple(e for op in a.ufl_operands for e in iter_expressions(op))
    elif isinstance(a, (Expr, BaseForm)):
        return (a,)
    raise ValueError(f"Not an UFL type: {type(a)}")



"""Basic algorithms for applying functions to subexpressions."""
# Copyright (C) 2014-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016

from ufl.core.expr import Expr
from ufl.corealg.multifunction import MultiFunction
from ufl.corealg.traversal import cutoff_unique_post_traversal, unique_post_traversal


def map_expr_dag(function, expression, compress=True, vcache=None, rcache=None):
    """Apply a function to each subexpression node in an expression DAG.

    If the same function is called multiple times in a transformation
    (as for example in apply_derivatives), then to reuse caches across
    the call, use the arguments vcache and rcache.

    Args:
        function: The function
        expression: An expression
        compress: If True (default), the output object from
            the function is cached in a dict and reused such that the
            resulting expression DAG does not contain duplicate objects
        vcache: Optional dict for caching results of intermediate
            transformations
        rcache: Optional dict for caching results for compression

    Returns:
        The result of the final function call
    """
    (result,) = map_expr_dags(
        function, [expression], compress=compress, vcache=vcache, rcache=rcache
    )
    return result


def map_expr_dags(function, expressions, compress=True, vcache=None, rcache=None):
    """Apply a function to each sub-expression node in an expression DAG.

    If *compress* is ``True`` (default) the output object from
    the function is cached in a ``dict`` and reused such that the
    resulting expression DAG does not contain duplicate objects.

    If the same function is called multiple times in a transformation
    (as for example in apply_derivatives), then to reuse caches across
    the call, use the arguments vcache and rcache.

    Args:
        function: The function
        expressions: An expression
        compress: If True (default), the output object from
            the function is cached in a dict and reused such that the
            resulting expression DAG does not contain duplicate objects
        vcache: Optional dict for caching results of intermediate transformations
        rcache: Optional dict for caching results for compression

    Returns:
        a list with the result of the final function call for each expression
    """
    # Temporary data structures
    # expr -> r = function(expr,...),  cache of intermediate results
    vcache = {} if vcache is None else vcache
    # r -> r,  cache of result objects for memory reuse
    rcache = {} if rcache is None else rcache

    # Build mapping typecode:bool, for which types to skip the subtree of
    if isinstance(function, MultiFunction):
        cutoff_types = function._is_cutoff_type
        handlers = function._handlers  # Optimization
    else:
        # Regular function: no skipping supported
        cutoff_types = [False] * Expr._ufl_num_typecodes_
        handlers = [function] * Expr._ufl_num_typecodes_

    # Create visited set here to share between traversal calls
    visited = set()

    # Pick faster traversal algorithm if we have no cutoffs
    if any(cutoff_types):

        def traversal(expression):
            return cutoff_unique_post_traversal(expression, cutoff_types, visited)
    else:

        def traversal(expression):
            return unique_post_traversal(expression, visited)

    for expression in expressions:
        # Iterate over all subexpression nodes, child before parent
        for v in traversal(expression):
            # Skip transformations on cache hit
            if v in vcache:
                continue

            # Cache miss: Get transformed operands, then apply transformation
            if cutoff_types[v._ufl_typecode_]:
                r = handlers[v._ufl_typecode_](v)
            else:
                r = handlers[v._ufl_typecode_](v, *[vcache[u] for u in v.ufl_operands])

            # Optionally check if r is in rcache, a memory optimization
            # to be able to keep representation of result compact
            if compress:
                r2 = rcache.get(r)
                if r2 is None:
                    # Cache miss: store in rcache
                    rcache[r] = r
                else:
                    # Cache hit: Use previously computed object r2,
                    # allowing r to be garbage collected as soon as possible
                    r = r2

            # Store result in cache
            vcache[v] = r

    return [vcache[expression] for expression in expressions]



from ufl import (
    CellVolume,
    Coefficient,
    Constant,
    FacetNormal,
    FunctionSpace,
    Identity,
    Mesh,
    SpatialCoordinate,
    TestFunction,
    VectorConstant,
    as_ufl,
    cos,
    derivative,
    diff,
    exp,
    grad,
    ln,
    sin,
    tan,
    triangle,
    variable,
    zero,
)
from ufl.algorithms.apply_derivatives import (
    GenericDerivativeRuleset,
    GradRuleset,
    apply_derivatives,
)
from ufl.algorithms.renumbering import renumber_indices
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2

# Note: the old tests in test_automatic_differentiation.py are a bit messy
#       but still cover many things that are not in here yet.


# FIXME: Write UNIT tests for all terminal derivatives!
# FIXME: Write UNIT tests for operator derivatives!


def test_apply_derivatives_doesnt_change_expression_without_derivatives():
    cell = triangle
    d = 2
    V0 = FiniteElement("Discontinuous Lagrange", cell, 0, (), identity_pullback, L2)
    V1 = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))
    v0_space = FunctionSpace(domain, V0)
    v1_space = FunctionSpace(domain, V1)

    # Literals
    z = zero((3, 2))
    one = as_ufl(1)
    two = as_ufl(2.0)
    ident = Identity(d)
    literals = [z, one, two, ident]

    # Geometry
    x = SpatialCoordinate(domain)
    n = FacetNormal(domain)
    volume = CellVolume(domain)
    geometry = [x, n, volume]

    # Arguments
    v0 = TestFunction(v0_space)
    v1 = TestFunction(v1_space)
    arguments = [v0, v1]

    # Coefficients
    f0 = Coefficient(v0_space)
    f1 = Coefficient(v1_space)
    coefficients = [f0, f1]

    # Expressions
    e0 = f0 + f1
    e1 = v0 * (f1 / 3 - f0**2)
    e2 = exp(sin(cos(tan(ln(x[0])))))
    expressions = [e0, e1, e2]

    # Check that all are unchanged
    for expr in literals + geometry + arguments + coefficients + expressions:
        # Note the use of "is" here instead of ==, this property
        # is important for efficiency and memory usage
        assert apply_derivatives(expr) is expr


def test_literal_derivatives_are_zero():
    cell = triangle
    d = 2

    # Literals
    one = as_ufl(1)
    two = as_ufl(2.0)
    ident = Identity(d)
    literals = [one, two, ident]

    # Generic ruleset handles literals directly:
    for lit in literals:
        for sh in [(), (d,), (d, d + 1)]:
            assert GenericDerivativeRuleset(sh)(lit) == zero(lit.ufl_shape + sh)

    # Variables
    v0 = variable(one)
    v1 = variable(zero((d,)))
    v2 = variable(ident)
    variables = [v0, v1, v2]

    # Test literals via apply_derivatives and variable ruleset:
    for lit in literals:
        for v in variables:
            assert apply_derivatives(diff(lit, v)) == zero(lit.ufl_shape + v.ufl_shape)

    V0 = FiniteElement("Discontinuous Lagrange", cell, 0, (), identity_pullback, L2)
    V1 = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))
    v0_space = FunctionSpace(domain, V0)
    v1_space = FunctionSpace(domain, V1)
    u0 = Coefficient(v0_space)
    u1 = Coefficient(v1_space)
    v0 = TestFunction(v0_space)
    v1 = TestFunction(v1_space)
    args = [(u0, v0), (u1, v1)]

    # Test literals via apply_derivatives and variable ruleset:
    for lit in literals:
        for u, v in args:
            assert apply_derivatives(derivative(lit, u, v)) == zero(lit.ufl_shape + v.ufl_shape)

    # Test grad ruleset directly since grad(literal) is invalid:
    assert GradRuleset(d)(one) == zero((d,))
    assert GradRuleset(d)(one) == zero((d,))


def test_grad_ruleset():
    cell = triangle
    d = 2

    V0 = FiniteElement("Discontinuous Lagrange", cell, 0, (), identity_pullback, L2)
    V1 = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V2 = FiniteElement("Lagrange", cell, 2, (), identity_pullback, H1)
    W0 = FiniteElement("Discontinuous Lagrange", cell, 0, (2,), identity_pullback, L2)
    W1 = FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1)
    W2 = FiniteElement("Lagrange", cell, 2, (d,), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", cell, 1, (d,), identity_pullback, H1))
    v0_space = FunctionSpace(domain, V0)
    v1_space = FunctionSpace(domain, V1)
    v2_space = FunctionSpace(domain, V2)
    w0_space = FunctionSpace(domain, W0)
    w1_space = FunctionSpace(domain, W1)
    w2_space = FunctionSpace(domain, W2)

    # Literals
    one = as_ufl(1)
    two = as_ufl(2.0)
    ident = Identity(d)

    # Geometry
    x = SpatialCoordinate(domain)
    n = FacetNormal(domain)
    volume = CellVolume(domain)

    # Arguments
    u0 = TestFunction(v0_space)
    u1 = TestFunction(v1_space)
    arguments = [u0, u1]

    # Coefficients
    r = Constant(domain)
    vr = VectorConstant(domain)
    f0 = Coefficient(v0_space)
    f1 = Coefficient(v1_space)
    f2 = Coefficient(v2_space)
    vf0 = Coefficient(w0_space)
    vf1 = Coefficient(w1_space)
    vf2 = Coefficient(w2_space)

    rules = GradRuleset(d)

    # Literals
    assert rules(one) == zero((d,))
    assert rules(two) == zero((d,))
    assert rules(ident) == zero((d, d, d))

    # Assumed piecewise constant geometry
    for g in [n, volume]:
        assert rules(g) == zero(g.ufl_shape + (d,))

    # Non-constant geometry
    assert rules(x) == ident

    # Arguments
    for u in arguments:
        assert rules(u) == grad(u)

    # Piecewise constant coefficients (Constant)
    assert rules(r) == zero((d,))
    assert rules(vr) == zero((d, d))
    assert rules(grad(r)) == zero((d, d))
    assert rules(grad(vr)) == zero((d, d, d))

    # Piecewise constant coefficients (DG0)
    assert rules(f0) == zero((d,))
    assert rules(vf0) == zero((d, d))
    assert rules(grad(f0)) == zero((d, d))
    assert rules(grad(vf0)) == zero((d, d, d))

    # Piecewise linear coefficients
    assert rules(f1) == grad(f1)
    assert rules(vf1) == grad(vf1)
    # assert rules(grad(f1)) == zero((d,d)) # TODO: Use degree to make this work
    # assert rules(grad(vf1)) == zero((d,d,d))

    # Piecewise quadratic coefficients
    assert rules(grad(f2)) == grad(grad(f2))
    assert rules(grad(vf2)) == grad(grad(vf2))

    # Indexed coefficients
    assert renumber_indices(apply_derivatives(grad(vf2[0]))) == renumber_indices(grad(vf2)[0, :])
    assert renumber_indices(apply_derivatives(grad(vf2[1])[0])) == renumber_indices(grad(vf2)[1, 0])

    # Grad of gradually more complex expressions
    assert apply_derivatives(grad(2 * f0)) == zero((d,))
    assert renumber_indices(apply_derivatives(grad(2 * f1))) == renumber_indices(2 * grad(f1))
    assert renumber_indices(apply_derivatives(grad(sin(f1)))) == renumber_indices(
        cos(f1) * grad(f1)
    )
    assert renumber_indices(apply_derivatives(grad(cos(f1)))) == renumber_indices(
        -sin(f1) * grad(f1)
    )


def test_variable_ruleset():
    pass


def test_gateaux_ruleset():
    pass



"""This module defines the Coefficient class and a number of related classes, including Constant."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008-2009.
# Modified by Massimiliano Leoni, 2016.
# Modified by Cecile Daversin-Catty, 2018.
# Modified by Ignacia Fierro-Piccardo 2023.

from ufl.argument import Argument
from ufl.core.terminal import FormArgument
from ufl.core.ufl_type import ufl_type
from ufl.duals import is_dual, is_primal
from ufl.form import BaseForm
from ufl.functionspace import AbstractFunctionSpace, MixedFunctionSpace
from ufl.split_functions import split
from ufl.utils.counted import Counted

# --- The Coefficient class represents a coefficient in a form ---


class BaseCoefficient(Counted):
    """UFL form argument type: Parent Representation of a form coefficient."""

    # Slots are disabled here because they cause trouble in PyDOLFIN
    # multiple inheritance pattern:
    # __slots__ = ("_count", "_ufl_function_space", "_repr", "_ufl_shape")
    _ufl_noslots_ = True
    __slots__ = ()
    _ufl_is_abstract_ = True

    def __getnewargs__(self):
        """Get new args."""
        return (self._ufl_function_space, self._count)

    def __init__(self, function_space, count=None):
        """Initalise."""
        Counted.__init__(self, count, Coefficient)

        if not isinstance(function_space, AbstractFunctionSpace):
            raise ValueError("Expecting a FunctionSpace.")

        self._ufl_function_space = function_space
        self._ufl_shape = function_space.value_shape

        self._repr = "BaseCoefficient(%s, %s)" % (repr(self._ufl_function_space), repr(self._count))

    @property
    def ufl_shape(self):
        """Return the associated UFL shape."""
        return self._ufl_shape

    def ufl_function_space(self):
        """Get the function space of this coefficient."""
        return self._ufl_function_space

    def ufl_domain(self):
        """Shortcut to get the domain of the function space of this coefficient."""
        return self._ufl_function_space.ufl_domain()

    def ufl_element(self):
        """Shortcut to get the finite element of the function space of this coefficient."""
        return self._ufl_function_space.ufl_element()

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        return self.ufl_element().is_cellwise_constant()

    def ufl_domains(self):
        """Return tuple of domains related to this terminal object."""
        return self._ufl_function_space.ufl_domains()

    def _ufl_signature_data_(self, renumbering):
        """Signature data.

        Signature data for form arguments depend on the global numbering
        of the form arguments and domains.
        """
        count = renumbering[self]
        fsdata = self._ufl_function_space._ufl_signature_data_(renumbering)
        return ("Coefficient", count, fsdata)

    def __str__(self):
        """Format as a string."""
        return f"w_{self._count}"

    def __repr__(self):
        """Representation."""
        return self._repr

    def __eq__(self, other):
        """Check equality."""
        if not isinstance(other, BaseCoefficient):
            return False
        if self is other:
            return True
        return self._count == other._count and self._ufl_function_space == other._ufl_function_space


@ufl_type()
class Cofunction(BaseCoefficient, BaseForm):
    """UFL form argument type: Representation of a form coefficient from a dual space."""

    __slots__ = (
        "_count",
        "_counted_class",
        "_arguments",
        "_coefficients",
        "_ufl_function_space",
        "ufl_operands",
        "_repr",
        "_ufl_shape",
        "_hash",
    )
    _primal = False
    _dual = True
    _ufl_is_terminal_ = True

    __eq__ = BaseForm.__eq__

    def __new__(cls, *args, **kw):
        """Create a new Cofunction."""
        if args[0] and is_primal(args[0]):
            raise ValueError(
                "ufl.Cofunction takes in a dual space. If you want to define a coefficient "
                "in the primal space you should use ufl.Coefficient."
            )
        return super().__new__(cls)

    def __init__(self, function_space, count=None):
        """Initialise."""
        BaseCoefficient.__init__(self, function_space, count)
        BaseForm.__init__(self)

        self.ufl_operands = ()
        self._hash = None
        self._repr = "Cofunction(%s, %s)" % (repr(self._ufl_function_space), repr(self._count))

    def equals(self, other):
        """Check equality."""
        if type(other) is not Cofunction:
            return False
        if self is other:
            return True
        return self._count == other._count and self._ufl_function_space == other._ufl_function_space

    def __hash__(self):
        """Hash."""
        return hash(("Cofunction", hash(self._ufl_function_space), self._count))

    def _analyze_form_arguments(self):
        """Analyze which Argument and Coefficient objects can be found in the form."""
        # Define canonical numbering of arguments and coefficients
        # Cofunctions have one argument in primal space as they map from V to R.
        self._arguments = (Argument(self._ufl_function_space.dual(), 0),)
        self._coefficients = (self,)


@ufl_type()
class Coefficient(FormArgument, BaseCoefficient):
    """UFL form argument type: Representation of a form coefficient."""

    _ufl_noslots_ = True
    _primal = True
    _dual = False

    __getnewargs__ = BaseCoefficient.__getnewargs__
    __str__ = BaseCoefficient.__str__
    _ufl_signature_data_ = BaseCoefficient._ufl_signature_data_

    def __new__(cls, *args, **kw):
        """Create a new Coefficient."""
        if args[0] and is_dual(args[0]):
            return Cofunction(*args, **kw)
        return super().__new__(cls)

    def __init__(self, function_space, count=None):
        """Initialise."""
        FormArgument.__init__(self)
        BaseCoefficient.__init__(self, function_space, count)

        self._repr = "Coefficient(%s, %s)" % (repr(self._ufl_function_space), repr(self._count))

    def ufl_domains(self):
        """Get the UFL domains."""
        return BaseCoefficient.ufl_domains(self)

    def __eq__(self, other):
        """Check equality."""
        if not isinstance(other, Coefficient):
            return False
        if self is other:
            return True
        return self._count == other._count and self._ufl_function_space == other._ufl_function_space

    def __repr__(self):
        """Representation."""
        return self._repr


# --- Helper functions for subfunctions on mixed elements ---


def Coefficients(function_space):
    """Create a Coefficient in a mixed space.

    Returns a tuple with the function components corresponding to the subelements.
    """
    if isinstance(function_space, MixedFunctionSpace):
        return [
            Coefficient(fs) if is_primal(fs) else Cofunction(fs)
            for fs in function_space.num_sub_spaces()
        ]
    else:
        return split(Coefficient(function_space))



"""Argument.

This module defines the class Argument and a number of related
classes (functions), including TestFunction and TrialFunction.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2008-2009.
# Modified by Massimiliano Leoni, 2016.
# Modified by Cecile Daversin-Catty, 2018.
# Modified by Ignacia Fierro-Piccardo 2023.

import numbers

from ufl.core.terminal import FormArgument
from ufl.core.ufl_type import ufl_type
from ufl.duals import is_dual, is_primal
from ufl.form import BaseForm
from ufl.functionspace import AbstractFunctionSpace, MixedFunctionSpace
from ufl.split_functions import split

# Export list for ufl.classes (TODO: not actually classes: drop? these are in ufl.*)
__all_classes__ = ["TestFunction", "TrialFunction", "TestFunctions", "TrialFunctions"]


# --- Class representing an argument (basis function) in a form ---


class BaseArgument(object):
    """UFL value: Representation of an argument to a form."""

    __slots__ = ()
    _ufl_is_abstract_ = True

    def __getnewargs__(self):
        """Get new args."""
        return (self._ufl_function_space, self._number, self._part)

    def __init__(self, function_space, number, part=None):
        """Initialise."""
        if not isinstance(function_space, AbstractFunctionSpace):
            raise ValueError("Expecting a FunctionSpace.")

        self._ufl_function_space = function_space
        self._ufl_shape = function_space.value_shape

        if not isinstance(number, numbers.Integral):
            raise ValueError(f"Expecting an int for number, not {number}")
        if part is not None and not isinstance(part, numbers.Integral):
            raise ValueError(f"Expecting None or an int for part, not {part}")
        self._number = number
        self._part = part

        self._repr = f"BaseArgument({self._ufl_function_space}, {self._number}, {self._part})"

    @property
    def ufl_shape(self):
        """Return the associated UFL shape."""
        return self._ufl_shape

    def ufl_function_space(self):
        """Get the function space of this Argument."""
        return self._ufl_function_space

    def ufl_domain(self):
        """Return the UFL domain."""
        return self._ufl_function_space.ufl_domain()

    def ufl_element(self):
        """Return The UFL element."""
        return self._ufl_function_space.ufl_element()

    def number(self):
        """Return the Argument number."""
        return self._number

    def part(self):
        """Return the part."""
        return self._part

    def is_cellwise_constant(self):
        """Return whether this expression is spatially constant over each cell."""
        # TODO: Should in principle do like with Coefficient,
        # but that may currently simplify away some arguments
        # we want to keep, or? See issue#13.
        # When we can annotate zero with arguments, we can change this.
        return False

    def ufl_domains(self):
        """Return UFL domains."""
        return self._ufl_function_space.ufl_domains()

    def _ufl_signature_data_(self, renumbering):
        """Signature data.

        Signature data for form arguments depend on the global numbering
        of the form arguments and domains.
        """
        fsdata = self._ufl_function_space._ufl_signature_data_(renumbering)
        return ("Argument", self._number, self._part, fsdata)

    def __str__(self):
        """Format as a string."""
        number = str(self._number)
        if len(number) == 1:
            s = "v_%s" % number
        else:
            s = "v_{%s}" % number
        if self._part is not None:
            part = str(self._part)
            if len(part) == 1:
                s = "%s^%s" % (s, part)
            else:
                s = "%s^{%s}" % (s, part)
        return s

    def __repr__(self):
        """Representation."""
        return self._repr

    def __eq__(self, other):
        """Check equality.

        Deliberately comparing exact type and not using isinstance here,
        meaning eventual subclasses must reimplement this function to work
        correctly, and instances of this class will compare not equal to
        instances of eventual subclasses. The overloading allows
        subclasses to distinguish between test and trial functions
        with a different non-ufl payload, such as dolfin FunctionSpace
        with different mesh. This is necessary because arguments with the
        same element and argument number are always equal from a pure ufl
        point of view, e.g. TestFunction(V1) == TestFunction(V2) if V1 and V2
        are the same ufl element but different dolfin function spaces.
        """
        return (
            type(self) is type(other)
            and self._number == other._number
            and self._part == other._part
            and self._ufl_function_space == other._ufl_function_space
        )


@ufl_type()
class Argument(FormArgument, BaseArgument):
    """UFL value: Representation of an argument to a form."""

    __slots__ = (
        "_ufl_function_space",
        "_ufl_shape",
        "_number",
        "_part",
        "_repr",
    )

    _primal = True
    _dual = False

    __getnewargs__ = BaseArgument.__getnewargs__
    __str__ = BaseArgument.__str__
    _ufl_signature_data_ = BaseArgument._ufl_signature_data_
    __eq__ = BaseArgument.__eq__

    def __new__(cls, *args, **kw):
        """Create new Argument."""
        if args[0] and is_dual(args[0]):
            return Coargument(*args, **kw)
        return super().__new__(cls)

    def __init__(self, function_space, number, part=None):
        """Initialise."""
        FormArgument.__init__(self)
        BaseArgument.__init__(self, function_space, number, part)

        self._repr = "Argument(%s, %s, %s)" % (
            repr(self._ufl_function_space),
            repr(self._number),
            repr(self._part),
        )

    def ufl_domains(self):
        """Return UFL domains."""
        return BaseArgument.ufl_domains(self)

    def __repr__(self):
        """Representation."""
        return self._repr


@ufl_type()
class Coargument(BaseForm, BaseArgument):
    """UFL value: Representation of an argument to a form in a dual space."""

    __slots__ = (
        "_ufl_function_space",
        "_ufl_shape",
        "_arguments",
        "_coefficients",
        "ufl_operands",
        "_number",
        "_part",
        "_repr",
        "_hash",
    )

    _primal = False
    _dual = True

    def __new__(cls, *args, **kw):
        """Create a new Coargument."""
        if args[0] and is_primal(args[0]):
            raise ValueError(
                "ufl.Coargument takes in a dual space! If you want to define an argument "
                "in the primal space you should use ufl.Argument."
            )
        return super().__new__(cls)

    def __init__(self, function_space, number, part=None):
        """Initialise."""
        BaseArgument.__init__(self, function_space, number, part)
        BaseForm.__init__(self)

        self.ufl_operands = ()
        self._hash = None
        self._repr = "Coargument(%s, %s, %s)" % (
            repr(self._ufl_function_space),
            repr(self._number),
            repr(self._part),
        )

    def arguments(self, outer_form=None):
        """Return all Argument objects found in form."""
        if self._arguments is None:
            self._analyze_form_arguments(outer_form=outer_form)
        return self._arguments

    def _analyze_form_arguments(self, outer_form=None):
        """Analyze which Argument and Coefficient objects can be found in the form."""
        # Define canonical numbering of arguments and coefficients
        self._coefficients = ()
        # Coarguments map from V* to V*, i.e. V* -> V*, or equivalently V* x V -> R.
        # So they have one argument in the primal space and one in the dual space.
        # However, when they are composed with linear forms with dual
        # arguments, such as BaseFormOperators, the primal argument is
        # discarded when analysing the argument as Coarguments.
        if not outer_form:
            self._arguments = (Argument(self.ufl_function_space().dual(), 0), self)
        else:
            self._arguments = (self,)

    def ufl_domain(self):
        """Return the UFL domain."""
        return BaseArgument.ufl_domain(self)

    def equals(self, other):
        """Check equality."""
        if type(other) is not Coargument:
            return False
        if self is other:
            return True
        return (
            self._ufl_function_space == other._ufl_function_space
            and self._number == other._number
            and self._part == other._part
        )

    def __hash__(self):
        """Hash."""
        return hash(("Coargument", hash(self._ufl_function_space), self._number, self._part))


# --- Helper functions for pretty syntax ---


def TestFunction(function_space, part=None):
    """UFL value: Create a test function argument to a form."""
    return Argument(function_space, 0, part)


def TrialFunction(function_space, part=None):
    """UFL value: Create a trial function argument to a form."""
    return Argument(function_space, 1, part)


# --- Helper functions for creating subfunctions on mixed elements ---


def Arguments(function_space, number):
    """Create an Argument in a mixed space.

    Returns a tuple with the function components corresponding to the subelements.
    """
    if isinstance(function_space, MixedFunctionSpace):
        return [
            Argument(function_space.ufl_sub_space(i), number, i)
            for i in range(function_space.num_sub_spaces())
        ]
    else:
        return split(Argument(function_space, number))


def TestFunctions(function_space):
    """Create a TestFunction in a mixed space.

    Returns a tuple with the function components corresponding to the
    subelements.
    """
    return Arguments(function_space, 0)


def TrialFunctions(function_space):
    """Create a TrialFunction in a mixed space.

    Returns a tuple with the function components corresponding to the
    subelements.
    """
    return Arguments(function_space, 1)



"""Pickle all the unit test forms from FFC 0.5.0"""

__author__ = "Anders Logg (logg@simula.no) et al."
__date__ = "2008-04-09 -- 2008-09-26"
__copyright__ = "Copyright (C) 2008 Anders Logg et al."
__license__ = "GNU GPL version 3 or any later version"

# Examples copied from the FFC demo directory, examples contributed
# by Johan Jansson, Kristian Oelgaard, Marie Rognes, and Garth Wells.

import pickle

from ufl import (
    Coefficient,
    Constant,
    Dx,
    FacetNormal,
    FunctionSpace,
    Identity,
    Mesh,
    TestFunction,
    TestFunctions,
    TrialFunction,
    TrialFunctions,
    VectorConstant,
    avg,
    curl,
    div,
    dot,
    dS,
    ds,
    dx,
    grad,
    i,
    inner,
    j,
    jump,
    lhs,
    rhs,
    sqrt,
    tetrahedron,
    triangle,
)
from ufl.algorithms import compute_form_data
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import contravariant_piola, covariant_piola, identity_pullback
from ufl.sobolevspace import H1, L2, HCurl, HDiv

p = pickle.HIGHEST_PROTOCOL


def testConstant():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    c = Constant(domain)
    d = VectorConstant(domain)

    a = c * dot(grad(v), grad(u)) * dx

    # FFC notation: L = dot(d, grad(v))*dx
    L = inner(d, grad(v)) * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testElasticity():
    element = FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    def eps(v):
        # FFC notation: return grad(v) + transp(grad(v))
        return grad(v) + (grad(v)).T

    # FFC notation: a = 0.25*dot(eps(v), eps(u))*dx
    a = 0.25 * inner(eps(v), eps(u)) * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)

    assert a.signature() == a_restore.signature()


def testEnergyNorm():
    element = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = Coefficient(space)
    a = (v * v + dot(grad(v), grad(v))) * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)

    assert a.signature() == a_restore.signature()


def testEquation():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    k = 0.1

    v = TestFunction(space)
    u = TrialFunction(space)
    u0 = Coefficient(space)

    F = v * (u - u0) * dx + k * dot(grad(v), grad(0.5 * (u0 + u))) * dx

    a = lhs(F)
    L = rhs(F)

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testFunctionOperators():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)
    g = Coefficient(space)

    # FFC notation: a = sqrt(1/modulus(1/f))*sqrt(g)*dot(grad(v), grad(u))*dx
    # + v*u*sqrt(f*g)*g*dx
    a = sqrt(1 / abs(1 / f)) * sqrt(g) * dot(grad(v), grad(u)) * dx + v * u * sqrt(f * g) * g * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)

    assert a.signature() == a_restore.signature()


def testHeat():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u1 = TrialFunction(space)
    u0 = Coefficient(space)
    c = Coefficient(space)
    f = Coefficient(space)
    k = Constant(domain)

    a = v * u1 * dx + k * c * dot(grad(v), grad(u1)) * dx
    L = v * u0 * dx + k * v * f * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testMass():
    element = FiniteElement("Lagrange", tetrahedron, 3, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    a = v * u * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)

    assert a.signature() == a_restore.signature()


def testMixedMixedElement():
    P3 = FiniteElement("Lagrange", triangle, 3, (), identity_pullback, H1)

    element = MixedElement([[P3, P3], [P3, P3]])

    element_pickle = pickle.dumps(element, p)
    element_restore = pickle.loads(element_pickle)

    assert element == element_restore


def testMixedPoisson():
    q = 1

    BDM = FiniteElement("Brezzi-Douglas-Marini", triangle, q, (2,), contravariant_piola, HDiv)
    DG = FiniteElement("Discontinuous Lagrange", triangle, q - 1, (), identity_pullback, L2)

    mixed_element = MixedElement([BDM, DG])
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    mixed_space = FunctionSpace(domain, mixed_element)
    dg_space = FunctionSpace(domain, DG)

    (tau, w) = TestFunctions(mixed_space)
    (sigma, u) = TrialFunctions(mixed_space)

    f = Coefficient(dg_space)

    a = (dot(tau, sigma) - div(tau) * u + w * div(sigma)) * dx
    L = w * f * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testNavierStokes():
    element = FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    w = Coefficient(space)

    # FFC notation: a = v[i]*w[j]*D(u[i], j)*dx
    a = v[i] * w[j] * Dx(u[i], j) * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)

    assert a.signature() == a_restore.signature()


def testNeumannProblem():
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)
    g = Coefficient(space)

    # FFC notation: a = dot(grad(v), grad(u))*dx
    a = inner(grad(v), grad(u)) * dx

    # FFC notation: L = dot(v, f)*dx + dot(v, g)*ds
    L = inner(v, f) * dx + inner(v, g) * ds

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testOptimization():
    element = FiniteElement("Lagrange", triangle, 3, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)

    a = dot(grad(v), grad(u)) * dx
    L = v * f * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testP5tet():
    element = FiniteElement("Lagrange", tetrahedron, 5, (), identity_pullback, H1)

    element_pickle = pickle.dumps(element, p)
    element_restore = pickle.loads(element_pickle)

    assert element == element_restore


def testP5tri():
    element = FiniteElement("Lagrange", triangle, 5, (), identity_pullback, H1)

    element_pickle = pickle.dumps(element, p)
    pickle.loads(element_pickle)


def testPoissonDG():
    element = FiniteElement("Discontinuous Lagrange", triangle, 1, (), identity_pullback, L2)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)

    n = FacetNormal(domain)

    # FFC notation: h = MeshSize(domain), not supported by UFL
    h = Constant(domain)

    gN = Coefficient(space)

    alpha = 4.0
    gamma = 8.0

    # FFC notation
    # a = dot(grad(v), grad(u))*dx \
    #    - dot(avg(grad(v)), jump(u, n))*dS \
    #    - dot(jump(v, n), avg(grad(u)))*dS \
    #    + alpha/h('+')*dot(jump(v, n), jump(u, n))*dS \
    #    - dot(grad(v), mult(u,n))*ds \
    #    - dot(mult(v,n), grad(u))*ds \
    #    + gamma/h*v*u*ds

    a = (
        inner(grad(v), grad(u)) * dx
        - inner(avg(grad(v)), jump(u, n)) * dS
        - inner(jump(v, n), avg(grad(u))) * dS
        + alpha / h("+") * dot(jump(v, n), jump(u, n)) * dS
        - inner(grad(v), u * n) * ds
        - inner(u * n, grad(u)) * ds
        + gamma / h * v * u * ds
    )

    L = v * f * dx + v * gN * ds

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testPoisson():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)

    # Note: inner() also works
    a = dot(grad(v), grad(u)) * dx
    L = v * f * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testPoissonSystem():
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)

    # FFC notation: a = dot(grad(v), grad(u))*dx
    a = inner(grad(v), grad(u)) * dx

    # FFC notation: L = dot(v, f)*dx
    L = inner(v, f) * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testQuadratureElement():
    element = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)

    # FFC notation:
    # QE = QuadratureElement(triangle, 3)
    # sig = VectorQuadratureElement(triangle, 3)

    QE = FiniteElement("Quadrature", triangle, 3, (), identity_pullback, L2)
    sig = FiniteElement("Quadrature", triangle, 3, (2,), identity_pullback, L2)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    qe_space = FunctionSpace(domain, QE)
    sig_space = FunctionSpace(domain, sig)

    v = TestFunction(space)
    u = TrialFunction(space)
    u0 = Coefficient(space)
    C = Coefficient(qe_space)
    sig0 = Coefficient(sig_space)
    f = Coefficient(space)

    a = v.dx(i) * C * u.dx(i) * dx + v.dx(i) * 2 * u0 * u * u0.dx(i) * dx
    L = v * f * dx - dot(grad(v), sig0) * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testStokes():
    # UFLException: Shape mismatch in sum.

    P2 = FiniteElement("Lagrange", triangle, 2, (2,), identity_pullback, H1)
    P1 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    TH = MixedElement([P2, P1])

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    th_space = FunctionSpace(domain, TH)
    p2_space = FunctionSpace(domain, P2)

    (v, q) = TestFunctions(th_space)
    (u, r) = TrialFunctions(th_space)

    f = Coefficient(p2_space)

    # FFC notation:
    # a = (dot(grad(v), grad(u)) - div(v)*p + q*div(u))*dx
    a = (inner(grad(v), grad(u)) - div(v) * r + q * div(u)) * dx

    L = dot(v, f) * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testSubDomain():
    element = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    f = Coefficient(space)

    M = f * dx(2) + f * ds(5)

    M_pickle = pickle.dumps(M, p)
    M_restore = pickle.loads(M_pickle)

    assert M.signature() == M_restore.signature()


def testSubDomains():
    element = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    a = (
        v * u * dx(0)
        + 10.0 * v * u * dx(1)
        + v * u * ds(0)
        + 2.0 * v * u * ds(1)
        + v("+") * u("+") * dS(0)
        + 4.3 * v("+") * u("+") * dS(1)
    )

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)

    assert a.signature() == a_restore.signature()


def testTensorWeightedPoisson():
    # FFC notation:
    # P1 = FiniteElement("Lagrange", triangle, 1)
    # P0 = FiniteElement("Discontinuous Lagrange", triangle, 0)
    #
    # v = TestFunction(P1)
    # u = TrialFunction(P1)
    # f = Coefficient(P1)
    #
    # c00 = Coefficient(P0)
    # c01 = Coefficient(P0)
    # c10 = Coefficient(P0)
    # c11 = Coefficient(P0)
    #
    # C = [[c00, c01], [c10, c11]]
    #
    # a = dot(grad(v), mult(C, grad(u)))*dx

    P1 = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    P0 = FiniteElement("Discontinuous Lagrange", triangle, 0, (2, 2), identity_pullback, L2)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    p1_space = FunctionSpace(domain, P1)
    p0_space = FunctionSpace(domain, P0)

    v = TestFunction(p1_space)
    u = TrialFunction(p1_space)
    C = Coefficient(p0_space)

    a = inner(grad(v), C * grad(u)) * dx

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)

    assert a.signature() == a_restore.signature()


def testVectorLaplaceGradCurl():
    def HodgeLaplaceGradCurl(space, fspace):
        (tau, v) = TestFunctions(space)
        (sigma, u) = TrialFunctions(space)
        f = Coefficient(fspace)

        # FFC notation: a = (dot(tau, sigma) - dot(grad(tau), u) + dot(v,
        # grad(sigma)) + dot(curl(v), curl(u)))*dx
        a = (
            inner(tau, sigma)
            - inner(grad(tau), u)
            + inner(v, grad(sigma))
            + inner(curl(v), curl(u))
        ) * dx

        # FFC notation: L = dot(v, f)*dx
        L = inner(v, f) * dx

        return [a, L]

    shape = tetrahedron
    order = 1

    GRAD = FiniteElement("Lagrange", shape, order, (), identity_pullback, H1)

    # FFC notation: CURL = FiniteElement("Nedelec", shape, order-1)
    CURL = FiniteElement("N1curl", shape, order, (3,), covariant_piola, HCurl)

    VectorLagrange = FiniteElement("Lagrange", shape, order + 1, (3,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", shape, 1, (3,), identity_pullback, H1))

    [a, L] = HodgeLaplaceGradCurl(
        FunctionSpace(domain, MixedElement([GRAD, CURL])), FunctionSpace(domain, VectorLagrange)
    )

    a_pickle = pickle.dumps(a, p)
    a_restore = pickle.loads(a_pickle)
    L_pickle = pickle.dumps(L, p)
    L_restore = pickle.loads(L_pickle)

    assert a.signature() == a_restore.signature()
    assert L.signature() == L_restore.signature()


def testIdentity():
    i = Identity(2)
    i_pickle = pickle.dumps(i, p)
    i_restore = pickle.loads(i_pickle)
    assert i == i_restore


def testFormData():
    element = FiniteElement("Lagrange", tetrahedron, 3, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    a = v * u * dx

    form_data = compute_form_data(a)

    form_data_pickle = pickle.dumps(form_data, p)
    form_data_restore = pickle.loads(form_data_pickle)

    assert str(form_data) == str(form_data_restore)



#
# Harmonic map demo using one mixed function u to represent x and y.
# Author: Martin Alnes
# Date: 2009-04-09
#
from ufl import Coefficient, FunctionSpace, Mesh, derivative, dot, dx, grad, inner, split, triangle
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

cell = triangle
X = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
Y = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
M = MixedElement([X, Y])
domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, M)

u = Coefficient(space)
x, y = split(u)

L = inner(grad(x), grad(x)) * dx + dot(x, x) * y * dx

F = derivative(L, u)
J = derivative(F, u)

forms = [L, F, J]



"""Test ExternalOperator object."""

__authors__ = "Nacime Bouziani"
__date__ = "2019-03-26"

import pytest

from ufl import (
    Action,
    Argument,
    Coargument,
    Coefficient,
    Constant,
    Form,
    FunctionSpace,
    Matrix,
    Mesh,
    TestFunction,
    TrialFunction,
    action,
    adjoint,
    cos,
    derivative,
    dx,
    inner,
    replace,
    sin,
    triangle,
)
from ufl.algorithms import expand_derivatives
from ufl.algorithms.apply_derivatives import apply_derivatives
from ufl.core.external_operator import ExternalOperator
from ufl.finiteelement import FiniteElement
from ufl.form import BaseForm
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


@pytest.fixture
def domain_2d():
    return Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))


@pytest.fixture
def V1(domain_2d):
    f1 = FiniteElement("CG", triangle, 1, (), identity_pullback, H1)
    return FunctionSpace(domain_2d, f1)


@pytest.fixture
def V2(domain_2d):
    f1 = FiniteElement("CG", triangle, 2, (), identity_pullback, H1)
    return FunctionSpace(domain_2d, f1)


@pytest.fixture
def V3(domain_2d):
    f1 = FiniteElement("CG", triangle, 3, (), identity_pullback, H1)
    return FunctionSpace(domain_2d, f1)


def test_properties(V1):
    u = Coefficient(V1, count=0)
    r = Coefficient(V1, count=1)

    e = ExternalOperator(u, r, function_space=V1)

    assert e.ufl_function_space() == V1
    assert e.ufl_operands[0] == u
    assert e.ufl_operands[1] == r
    assert e.derivatives == (0, 0)
    assert e.ufl_shape == ()

    e2 = ExternalOperator(u, r, function_space=V1, derivatives=(3, 4))
    assert e2.derivatives == (3, 4)
    assert e2.ufl_shape == ()

    # Test __str__
    s = Coefficient(V1, count=2)
    t = Coefficient(V1, count=3)
    v0 = Argument(V1, 0)
    v1 = Argument(V1, 1)

    e = ExternalOperator(u, function_space=V1)
    assert str(e) == "e(w_0; v_0)"

    e = ExternalOperator(u, function_space=V1, derivatives=(1,))
    assert str(e) == "∂e(w_0; v_0)/∂o1"

    e = ExternalOperator(
        u, r, 2 * s, t, function_space=V1, derivatives=(1, 0, 1, 2), argument_slots=(v0, v1)
    )
    assert str(e) == "∂e(w_0, w_1, 2 * w_2, w_3; v_1, v_0)/∂o1∂o3∂o4∂o4"


def test_form(V1, V2):
    u = Coefficient(V1)
    m = Coefficient(V1)
    u_hat = TrialFunction(V1)
    v = TestFunction(V1)

    # F = N * v * dx
    N = ExternalOperator(u, m, function_space=V2)
    F = N * v * dx
    actual = derivative(F, u, u_hat)

    (vstar,) = N.arguments()
    Nhat = TrialFunction(N.ufl_function_space())

    dNdu = N._ufl_expr_reconstruct_(u, m, derivatives=(1, 0), argument_slots=(vstar, u_hat))
    dFdN = Nhat * v * dx
    expected = Action(dFdN, dNdu)

    assert apply_derivatives(actual) == expected

    # F = N * u * v * dx
    N = ExternalOperator(u, m, function_space=V1)
    F = N * u * v * dx
    actual = derivative(F, u, u_hat)

    (vstar,) = N.arguments()
    Nhat = TrialFunction(N.ufl_function_space())

    dNdu = N._ufl_expr_reconstruct_(u, m, derivatives=(1, 0), argument_slots=(vstar, u_hat))
    dFdu_partial = N * u_hat * v * dx
    dFdN = Nhat * u * v * dx
    expected = dFdu_partial + Action(dFdN, dNdu)
    assert apply_derivatives(actual) == expected


def test_differentiation_procedure_action(V1, V2):
    s = Coefficient(V1)
    u = Coefficient(V2)
    m = Coefficient(V2)

    # External operators
    N1 = ExternalOperator(u, m, function_space=V1)
    N2 = ExternalOperator(cos(s), function_space=V1)

    # Check arguments and argument slots
    assert len(N1.arguments()) == 1
    assert len(N2.arguments()) == 1
    assert N1.arguments() == N1.argument_slots()
    assert N2.arguments() == N2.argument_slots()

    # Check coefficients
    assert N1.coefficients() == (u, m)
    assert N2.coefficients() == (s,)

    # Get v*
    (vstar_N1,) = N1.arguments()
    (vstar_N2,) = N2.arguments()
    assert vstar_N1.ufl_function_space().dual() == V1
    assert vstar_N2.ufl_function_space().dual() == V1

    u_hat = Argument(V1, 1)
    s_hat = Argument(V2, 1)
    w = Coefficient(V1)
    r = Coefficient(V2)

    # Bilinear forms
    a1 = inner(N1, m) * dx
    Ja1 = derivative(a1, u, u_hat)
    Ja1 = expand_derivatives(Ja1)

    a2 = inner(N2, m) * dx
    Ja2 = derivative(a2, s, s_hat)
    Ja2 = expand_derivatives(Ja2)

    # Get external operators
    assert isinstance(Ja1, Action)
    dN1du = Ja1.right()
    dN1du_action = Action(dN1du, w)

    assert isinstance(Ja2, Action)
    dN2du = Ja2.right()
    dN2du_action = Action(dN2du, r)

    # Check shape
    assert dN1du.ufl_shape == ()
    assert dN2du.ufl_shape == ()

    # Get v*s
    vstar_dN1du, _ = dN1du.arguments()
    vstar_dN2du, _ = dN2du.arguments()
    assert vstar_dN1du.ufl_function_space().dual() == V1  # shape: (2,)
    assert vstar_dN2du.ufl_function_space().dual() == V1  # shape: (2,)

    # Check derivatives
    assert dN1du.derivatives == (1, 0)
    assert dN2du.derivatives == (1,)

    # Check arguments
    assert dN1du.arguments() == (vstar_dN1du, u_hat)
    assert dN1du_action.arguments() == (vstar_dN1du,)

    assert dN2du.arguments() == (vstar_dN2du, s_hat)
    assert dN2du_action.arguments() == (vstar_dN2du,)

    # Check argument slots
    assert dN1du.argument_slots() == (vstar_dN1du, u_hat)
    assert dN2du.argument_slots() == (vstar_dN2du, -sin(s) * s_hat)


def test_extractions(domain_2d, V1):
    from ufl.algorithms.analysis import (
        extract_arguments,
        extract_arguments_and_coefficients,
        extract_base_form_operators,
        extract_coefficients,
        extract_constants,
    )

    u = Coefficient(V1)
    c = Constant(domain_2d)

    e = ExternalOperator(u, c, function_space=V1)
    (vstar_e,) = e.arguments()

    assert extract_coefficients(e) == [u]
    assert extract_arguments(e) == [vstar_e]
    assert extract_arguments_and_coefficients(e) == ([vstar_e], [u])
    assert extract_constants(e) == [c]
    assert extract_base_form_operators(e) == [e]

    F = e * dx

    assert extract_coefficients(F) == [u]
    assert extract_arguments(e) == [vstar_e]
    assert extract_arguments_and_coefficients(e) == ([vstar_e], [u])
    assert extract_constants(F) == [c]
    assert F.base_form_operators() == (e,)

    u_hat = Argument(V1, 1)
    e = ExternalOperator(u, function_space=V1, derivatives=(1,), argument_slots=(vstar_e, u_hat))

    assert extract_coefficients(e) == [u]
    assert extract_arguments(e) == [vstar_e, u_hat]
    assert extract_arguments_and_coefficients(e) == ([vstar_e, u_hat], [u])
    assert extract_base_form_operators(e) == [e]

    F = e * dx

    assert extract_coefficients(F) == [u]
    assert extract_arguments(e) == [vstar_e, u_hat]
    assert extract_arguments_and_coefficients(e) == ([vstar_e, u_hat], [u])
    assert F.base_form_operators() == (e,)

    w = Coefficient(V1)
    e2 = ExternalOperator(w, e, function_space=V1)
    (vstar_e2,) = e2.arguments()

    assert extract_coefficients(e2) == [u, w]
    assert extract_arguments(e2) == [vstar_e2, u_hat]
    assert extract_arguments_and_coefficients(e2) == ([vstar_e2, u_hat], [u, w])
    assert extract_base_form_operators(e2) == [e, e2]

    F = e2 * dx

    assert extract_coefficients(e2) == [u, w]
    assert extract_arguments(e2) == [vstar_e2, u_hat]
    assert extract_arguments_and_coefficients(e2) == ([vstar_e2, u_hat], [u, w])
    assert F.base_form_operators() == (e, e2)


def get_external_operators(form_base):
    if isinstance(form_base, ExternalOperator):
        return (form_base,)
    elif isinstance(form_base, BaseForm):
        return form_base.base_form_operators()
    else:
        raise ValueError("Expecting BaseForm argument!")


def test_adjoint_action_jacobian(V1, V2, V3):
    u = Coefficient(V1)
    m = Coefficient(V2)

    # N(u, m; v*)
    N = ExternalOperator(u, m, function_space=V3)

    # Arguments for the Gateaux-derivative
    def u_hat(number):
        return Argument(V1, number)  # V1: degree 1 # dFdu.arguments()[-1]

    def m_hat(number):
        return Argument(V2, number)  # V2: degree 2 # dFdm.arguments()[-1]

    def vstar_N(number):
        return Argument(V3.dual(), number)  # V3: degree 3

    # Coefficients for the action
    w = Coefficient(V1)  # for u
    p = Coefficient(V2)  # for m

    v2 = TestFunction(V2)
    v3 = TestFunction(V3)
    form_base_expressions = (N * dx, N * v2 * dx, N * v3 * dx)  # , N)

    for F in form_base_expressions:
        # Get test function
        v_F = F.arguments() if isinstance(F, Form) else ()
        # If we have a 0-form with an ExternalOperator: e.g. F = N * dx
        # => F.arguments() = (), because of form composition.
        # But we still need to make arguments with number 1 (i.e. n_arg = 1)
        # since at the external operator level, argument numbering is based on
        # the external operator arguments and not on the outer form arguments.
        n_arg = len(v_F) if len(v_F) else 1
        assert n_arg < 2

        # Differentiate
        dFdu = expand_derivatives(derivative(F, u, u_hat(n_arg)))
        dFdm = expand_derivatives(derivative(F, m, m_hat(n_arg)))

        assert dFdu.arguments() == v_F + (u_hat(n_arg),)
        assert dFdm.arguments() == v_F + (m_hat(n_arg),)

        assert isinstance(dFdu, Action)

        # dNdu(u, m; u_hat, v*)
        dNdu = dFdu.right()
        # dNdm(u, m; m_hat, v*)
        dNdm = dFdm.right()

        assert dNdu.derivatives == (1, 0)
        assert dNdm.derivatives == (0, 1)
        assert dNdu.arguments() == (vstar_N(0), u_hat(n_arg))
        assert dNdm.arguments() == (vstar_N(0), m_hat(n_arg))
        assert dNdu.argument_slots() == dNdu.arguments()
        assert dNdm.argument_slots() == dNdm.arguments()

        # Action
        action_dFdu = action(dFdu, w)
        action_dFdm = action(dFdm, p)

        assert action_dFdu.arguments() == v_F + ()
        assert action_dFdm.arguments() == v_F + ()

        # If we have 2 arguments
        if len(v_F):
            # Adjoint
            dFdu_adj = adjoint(dFdu)
            dFdm_adj = adjoint(dFdm)

            V = v_F[0].ufl_function_space()
            assert dFdu_adj.arguments() == (TestFunction(V1), TrialFunction(V))
            assert dFdm_adj.arguments() == (TestFunction(V2), TrialFunction(V))

            # Action of the adjoint
            q = Coefficient(V)
            action_dFdu_adj = action(dFdu_adj, q)
            action_dFdm_adj = action(dFdm_adj, q)

            assert action_dFdu_adj.arguments() == (TestFunction(V1),)
            assert action_dFdm_adj.arguments() == (TestFunction(V2),)


def test_multiple_external_operators(V1, V2):
    u = Coefficient(V1)
    m = Coefficient(V1)
    w = Coefficient(V2)

    v = TestFunction(V1)
    v_hat = TrialFunction(V1)
    w_hat = TrialFunction(V2)

    # N1(u, m; v*)
    N1 = ExternalOperator(u, m, function_space=V1)

    # N2(w; v*)
    N2 = ExternalOperator(w, function_space=V2)

    # N3(u; v*)
    N3 = ExternalOperator(u, function_space=V1)

    # N4(N1, u; v*)
    N4 = ExternalOperator(N1, u, function_space=V1)

    # N5(N4(N1, u); v*)
    N5 = ExternalOperator(N4, u, function_space=V1)

    # --- F = < N1(u, m; v*), v > + <N2(w; v*), v> + <N3(u; v*), v> --- #

    F = (inner(N1, v) + inner(N2, v) + inner(N3, v)) * dx

    # dFdu = Action(dFdN1, dN1du) + Action(dFdN3, dN3du)
    dFdu = expand_derivatives(derivative(F, u))
    dFdN1 = inner(v_hat, v) * dx
    dFdN2 = inner(w_hat, v) * dx
    dFdN3 = inner(v_hat, v) * dx
    dN1du = N1._ufl_expr_reconstruct_(
        u, m, derivatives=(1, 0), argument_slots=N1.arguments() + (v_hat,)
    )
    dN3du = N3._ufl_expr_reconstruct_(u, derivatives=(1,), argument_slots=N3.arguments() + (v_hat,))

    assert dFdu == Action(dFdN1, dN1du) + Action(dFdN3, dN3du)

    # dFdm = Action(dFdN1, dN1dm)
    dFdm = expand_derivatives(derivative(F, m))
    dN1dm = N1._ufl_expr_reconstruct_(
        u, m, derivatives=(0, 1), argument_slots=N1.arguments() + (v_hat,)
    )

    assert dFdm == Action(dFdN1, dN1dm)

    # dFdw = Action(dFdN2, dN2dw)
    dFdw = expand_derivatives(derivative(F, w))
    dN2dw = N2._ufl_expr_reconstruct_(w, derivatives=(1,), argument_slots=N2.arguments() + (w_hat,))

    assert dFdw == Action(dFdN2, dN2dw)

    # --- F = < N4(N1(u, m), u; v*), v > --- #

    F = inner(N4, v) * dx

    # dFdu = ∂F/∂u + Action(∂F/∂N1, dN1/du) + Action(∂F/∂N4, dN4/du)
    #      = Action(∂F/∂N4, dN4/du), since ∂F/∂u = 0 and ∂F/∂N1 = 0
    #
    # In addition, we have:
    # dN4/du = ∂N4/∂u + Action(∂N4/∂N1, dN1/du)
    #
    # Using the fact that Action is distributive, we have:
    #
    # dFdu = Action(∂F/∂N4, ∂N4/∂u) +
    #         Action(∂F/∂N4, Action(∂N4/∂N1, dN1/du))
    dFdu = expand_derivatives(derivative(F, u))
    dFdN4_partial = inner(v_hat, v) * dx
    dN4dN1_partial = N4._ufl_expr_reconstruct_(
        N1, u, derivatives=(1, 0), argument_slots=N4.arguments() + (v_hat,)
    )
    dN4du_partial = N4._ufl_expr_reconstruct_(
        N1, u, derivatives=(0, 1), argument_slots=N4.arguments() + (v_hat,)
    )

    assert dFdu == Action(dFdN4_partial, Action(dN4dN1_partial, dN1du)) + Action(
        dFdN4_partial, dN4du_partial
    )

    # dFdm = Action(∂F/∂N4, Action(∂N4/∂N1, dN1/dm))
    dFdm = expand_derivatives(derivative(F, m))

    assert dFdm == Action(dFdN4_partial, Action(dN4dN1_partial, dN1dm))

    # --- F = < N1(u, m; v*), v > + <N2(w; v*), v> + <N3(u; v*), v> + <
    # N4(N1(u, m), u; v*), v > --- #

    F = (inner(N1, v) + inner(N2, v) + inner(N3, v) + inner(N4, v)) * dx

    dFdu = expand_derivatives(derivative(F, u))
    assert dFdu == Action(dFdN1, dN1du) + Action(dFdN3, dN3du) + Action(
        dFdN4_partial, Action(dN4dN1_partial, dN1du)
    ) + Action(dFdN4_partial, dN4du_partial)

    dFdm = expand_derivatives(derivative(F, m))
    assert dFdm == Action(dFdN1, dN1dm) + Action(dFdN4_partial, Action(dN4dN1_partial, dN1dm))

    dFdw = expand_derivatives(derivative(F, w))
    assert dFdw == Action(dFdN2, dN2dw)

    # --- F = < N5(N4(N1(u, m), u), u; v*), v > + < N1(u, m; v*), v > +
    # < u * N5(N4(N1(u, m), u), u; v*), v >--- #

    F = (inner(N5, v) + inner(N1, v) + inner(u * N5, v)) * dx

    # dFdu = ∂F/∂u + Action(∂F/∂N1, dN1/du) + Action(∂F/∂N4, dN4/du) + Action(∂F/∂N5, dN5/du)
    #
    # where:
    #  - ∂F/∂u = inner(w * N5, v) * dx
    #  - ∂F/∂N1 = inner(w, v) * dx
    #  - ∂F/∂N5 = inner(w, v) * dx + inner(u * w, v) * dx
    #  - ∂F/∂N4 = 0
    #  - dN5/du = ∂N5/∂u + Action(∂N5/∂N4, dN4/du)
    #           = ∂N5/∂u + Action(∂N5/∂N4, ∂N4/∂u) + Action(∂N5/∂N4, Action(∂N4/∂N1, dN1/du))
    # with w = TrialFunction(V1)
    w = TrialFunction(V1)
    dFdu_partial = inner(w * N5, v) * dx
    dFdN1_partial = inner(w, v) * dx
    dFdN5_partial = (inner(w, v) + inner(u * w, v)) * dx
    dN5dN4_partial = N5._ufl_expr_reconstruct_(
        N4, u, derivatives=(1, 0), argument_slots=N4.arguments() + (w,)
    )
    dN5du_partial = N5._ufl_expr_reconstruct_(
        N4, u, derivatives=(0, 1), argument_slots=N4.arguments() + (w,)
    )
    dN5du = (
        Action(dN5dN4_partial, Action(dN4dN1_partial, dN1du))
        + Action(dN5dN4_partial, dN4du_partial)
        + dN5du_partial
    )

    dFdu = expand_derivatives(derivative(F, u))
    assert dFdu == dFdu_partial + Action(dFdN1_partial, dN1du) + Action(dFdN5_partial, dN5du)


def test_replace(V1):
    u = Coefficient(V1, count=0)
    N = ExternalOperator(u, function_space=V1)

    # dN(u; uhat, v*)
    dN = expand_derivatives(derivative(N, u))
    vstar, uhat = dN.arguments()
    assert isinstance(vstar, Coargument)

    # Replace v* by a Form
    v = TestFunction(V1)
    F = inner(u, v) * dx
    G = replace(dN, {vstar: F})

    dN_replaced = dN._ufl_expr_reconstruct_(u, argument_slots=(F, uhat))
    assert G == dN_replaced

    # Replace v* by an Action
    M = Matrix(V1, V1)
    A = Action(M, u)
    G = replace(dN, {vstar: A})

    dN_replaced = dN._ufl_expr_reconstruct_(u, argument_slots=(A, uhat))
    assert G == dN_replaced



"""Expr operators.

This module attaches special functions to Expr.
This way we avoid circular dependencies between e.g.
Sum and its superclass Expr.
"""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016.

import numbers

from ufl.algebra import Abs, Division, Power, Product, Sum
from ufl.conditional import GE, GT, LE, LT
from ufl.constantvalue import Zero, as_ufl
from ufl.core.expr import Expr
from ufl.core.multiindex import Index, MultiIndex, indices
from ufl.differentiation import Grad
from ufl.exprequals import expr_equals
from ufl.index_combination_utils import create_slice_indices, merge_overlapping_indices
from ufl.indexed import Indexed
from ufl.indexsum import IndexSum
from ufl.restriction import NegativeRestricted, PositiveRestricted
from ufl.tensoralgebra import Inner, Transposed
from ufl.tensors import ComponentTensor, as_tensor
from ufl.utils.stacks import StackDict

# --- Boolean operators ---


def _le(left, right):
    """A boolean expresion (left <= right) for use with conditional."""
    return LE(left, right)


def _ge(left, right):
    """A boolean expresion (left >= right) for use with conditional."""
    return GE(left, right)


def _lt(left, right):
    """A boolean expresion (left < right) for use with conditional."""
    return LT(left, right)


def _gt(left, right):
    """A boolean expresion (left > right) for use with conditional."""
    return GT(left, right)


# '==' needs to implement comparison of expression representations for
# use in hashmaps (dict and set), but the others can be overloaded in
# the language.  It is possible that we can overload eq as well, but
# we'll need to fix some issues first and also check for a possible
# significant performance hit with compilation of complex
# forms. Replacing a==b with equiv(a,b) all over the code could be one
# way to reduce such a performance hit, but we cannot do anything
# about dict and set calling __eq__...
Expr.__eq__ = expr_equals


# != is used at least by tests, possibly in code as well, and must
# mean the opposite of ==, i.e. when evaluated as bool it must mean
# 'not equal representation'.
def _ne(self, other):
    return not self.__eq__(other)


Expr.__ne__ = _ne
Expr.__lt__ = _lt
Expr.__gt__ = _gt
Expr.__le__ = _le
Expr.__ge__ = _ge

# Python operators 'and'/'or' cannot be overloaded, and bitwise
# operators &/| don't have the right precedence levels
# Expr.__and__ = _and
# Expr.__or__ = _or


def _as_tensor(self, indices):
    """A^indices := as_tensor(A, indices)."""
    if not isinstance(indices, tuple):
        raise ValueError(
            "Expecting a tuple of Index objects to A^indices := as_tensor(A, indices)."
        )
    if not all(isinstance(i, Index) for i in indices):
        raise ValueError(
            "Expecting a tuple of Index objects to A^indices := as_tensor(A, indices)."
        )
    return as_tensor(self, indices)


Expr.__xor__ = _as_tensor


# --- Helper functions for product handling ---


def _mult(a, b):
    """Multiply."""
    # Discover repeated indices, which results in index sums
    afi = a.ufl_free_indices
    bfi = b.ufl_free_indices
    afid = a.ufl_index_dimensions
    bfid = b.ufl_index_dimensions
    fi, fid, ri, rid = merge_overlapping_indices(afi, afid, bfi, bfid)

    # Pick out valid non-scalar products here (dot products):
    # - matrix-matrix (A*B, M*grad(u)) => A . B
    # - matrix-vector (A*v) => A . v
    s1, s2 = a.ufl_shape, b.ufl_shape
    r1, r2 = len(s1), len(s2)

    if r1 == 0 and r2 == 0:
        # Create scalar product
        p = Product(a, b)
        ti = ()

    elif r1 == 0 or r2 == 0:
        # Scalar - tensor product
        if r2 == 0:
            a, b = b, a

        # Check for zero, simplifying early if possible
        if isinstance(a, Zero) or isinstance(b, Zero):
            shape = s1 or s2
            return Zero(shape, fi, fid)

        # Repeated indices are allowed, like in:
        # v[i]*M[i,:]

        # Apply product to scalar components
        ti = indices(len(b.ufl_shape))
        p = Product(a, b[ti])

    elif r1 == 2 and r2 in (1, 2):  # Matrix-matrix or matrix-vector
        if ri:
            raise ValueError("Not expecting repeated indices in non-scalar product.")

        # Check for zero, simplifying early if possible
        if isinstance(a, Zero) or isinstance(b, Zero):
            shape = s1[:-1] + s2[1:]
            return Zero(shape, fi, fid)

        # Return dot product in index notation
        ai = indices(len(a.ufl_shape) - 1)
        bi = indices(len(b.ufl_shape) - 1)
        k = indices(1)

        p = a[ai + k] * b[k + bi]
        ti = ai + bi

    else:
        raise ValueError(f"Invalid ranks {r1} and {r2} in product.")

    # TODO: I think applying as_tensor after index sums results in
    # cleaner expression graphs.
    # Wrap as tensor again
    if ti:
        p = as_tensor(p, ti)

    # If any repeated indices were found, apply implicit summation
    # over those
    for i in ri:
        mi = MultiIndex((Index(count=i),))
        p = IndexSum(p, mi)

    return p


# --- Extend Expr with algebraic operators ---

_valid_types = (Expr, numbers.Real, numbers.Integral, numbers.Complex)


def _mul(self, o):
    """Multiply."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    o = as_ufl(o)
    return _mult(self, o)


Expr.__mul__ = _mul


def _rmul(self, o):
    """Multiply."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    o = as_ufl(o)
    return _mult(o, self)


Expr.__rmul__ = _rmul


def _add(self, o):
    """Add."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    return Sum(self, o)


Expr.__add__ = _add


def _radd(self, o):
    """Add."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    if isinstance(o, numbers.Number) and o == 0:
        # Allow adding scalar int 0 as a no-op, even for shaped self,
        # needed for sum([a,b])
        return self
    return Sum(o, self)


Expr.__radd__ = _radd


def _sub(self, o):
    """Subtract."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    return Sum(self, -o)


Expr.__sub__ = _sub


def _rsub(self, o):
    """Subtract."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    return Sum(o, -self)


Expr.__rsub__ = _rsub


def _div(self, o):
    """Divide."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    sh = self.ufl_shape
    if sh:
        ii = indices(len(sh))
        d = Division(self[ii], o)
        return as_tensor(d, ii)
    return Division(self, o)


Expr.__div__ = _div
Expr.__truediv__ = _div


def _rdiv(self, o):
    """Divide."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    return Division(o, self)


Expr.__rdiv__ = _rdiv
Expr.__rtruediv__ = _rdiv


def _pow(self, o):
    """Raise to a power."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    if o == 2 and self.ufl_shape:
        return Inner(self, self)
    return Power(self, o)


Expr.__pow__ = _pow


def _rpow(self, o):
    """Raise to a power."""
    if not isinstance(o, _valid_types):
        return NotImplemented
    return Power(o, self)


Expr.__rpow__ = _rpow


# TODO: Add Negated class for this? Might simplify reductions in Add.
def _neg(self):
    """Negate."""
    return -1 * self


Expr.__neg__ = _neg


def _abs(self):
    """Absolute value."""
    return Abs(self)


Expr.__abs__ = _abs


# --- Extend Expr with restiction operators a("+"), a("-") ---


def _restrict(self, side):
    """Restrict."""
    if side == "+":
        return PositiveRestricted(self)
    if side == "-":
        return NegativeRestricted(self)
    raise ValueError(f"Invalid side '{side}' in restriction operator.")


def _eval(self, coord, mapping=None, component=()):
    """Evaluate.

    Evaluate expression at this particular coordinate, with provided
    values for other terminals in mapping.
    """
    # Evaluate derivatives first
    from ufl.algorithms import expand_derivatives

    f = expand_derivatives(self)

    # Evaluate recursively
    if mapping is None:
        mapping = {}
    index_values = StackDict()
    return f.evaluate(coord, mapping, component, index_values)


def _call(self, arg, mapping=None, component=()):
    """Take the restriction or evaluate depending on argument."""
    if arg in ("+", "-"):
        if mapping is not None:
            raise ValueError("Not expecting a mapping when taking restriction.")
        return _restrict(self, arg)
    else:
        return _eval(self, arg, mapping, component)


Expr.__call__ = _call


# --- Extend Expr with the transpose operation A.T ---


def _transpose(self):
    """Transpose a rank-2 tensor expression.

    For more general transpose operations of higher order tensor
    expressions, use indexing and Tensor.
    """
    return Transposed(self)


Expr.T = property(_transpose)


# --- Extend Expr with indexing operator a[i] ---


def _getitem(self, component):
    """Get an item."""
    # Treat component consistently as tuple below
    if not isinstance(component, tuple):
        component = (component,)

    shape = self.ufl_shape

    # Analyse slices (:) and Ellipsis (...)
    all_indices, slice_indices, repeated_indices = create_slice_indices(
        component, shape, self.ufl_free_indices
    )

    # Check that we have the right number of indices for a tensor with
    # this shape
    if len(shape) != len(all_indices):
        raise ValueError(
            f"Invalid number of indices {len(all_indices)} for expression of rank {len(shape)}."
        )

    # Special case for simplifying foo[...] => foo, foo[:] => foo or
    # similar
    if len(slice_indices) == len(all_indices):
        return self

    # Special case for simplifying as_tensor(ai,(i,))[i] => ai
    if isinstance(self, ComponentTensor):
        if all_indices == self.indices().indices():
            return self.ufl_operands[0]

    # Apply all indices to index self, yielding a scalar valued
    # expression
    mi = MultiIndex(all_indices)
    a = Indexed(self, mi)

    # TODO: I think applying as_tensor after index sums results in
    # cleaner expression graphs.

    # If the Ellipsis or any slices were found, wrap as tensor valued
    # with the slice indices created at the top here
    if slice_indices:
        a = as_tensor(a, slice_indices)

    # If any repeated indices were found, apply implicit summation
    # over those
    for i in repeated_indices:
        mi = MultiIndex((i,))
        a = IndexSum(a, mi)

    # Check for zero (last so we can get indices etc from a, could
    # possibly be done faster by checking early instead)
    if isinstance(self, Zero):
        shape = a.ufl_shape
        fi = a.ufl_free_indices
        fid = a.ufl_index_dimensions
        a = Zero(shape, fi, fid)

    return a


Expr.__getitem__ = _getitem


# --- Extend Expr with spatial differentiation operator a.dx(i) ---


def _dx(self, *ii):
    """Return the partial derivative with respect to spatial variable number *ii*."""
    d = self
    # Unwrap ii to allow .dx(i,j) and .dx((i,j))
    if len(ii) == 1 and isinstance(ii[0], tuple):
        ii = ii[0]
    # Apply all derivatives
    for i in ii:
        d = Grad(d)

    # Take all components, applying repeated index sums in the [] operation
    return d.__getitem__((Ellipsis,) + ii)


Expr.dx = _dx



import numpy as np

from ufl import Cell, Coefficient, FunctionSpace, Mesh, as_tensor, as_vector, dx, indices, triangle
from ufl.algorithms.renumbering import renumber_indices
from ufl.classes import Jacobian, JacobianDeterminant, JacobianInverse, ReferenceValue
from ufl.finiteelement import FiniteElement, MixedElement, SymmetricElement
from ufl.pullback import (
    contravariant_piola,
    covariant_piola,
    double_contravariant_piola,
    double_covariant_piola,
    identity_pullback,
    l2_piola,
)
from ufl.sobolevspace import H1, L2, HCurl, HDiv, HDivDiv, HEin


def check_single_function_pullback(g, mappings):
    expected = mappings[g]
    actual = g.ufl_element().pullback.apply(ReferenceValue(g))
    assert expected.ufl_shape == actual.ufl_shape
    for idx in np.ndindex(actual.ufl_shape):
        rexp = renumber_indices(expected[idx])
        ract = renumber_indices(actual[idx])
        if not rexp == ract:
            print()
            print("In check_single_function_pullback:")
            print("input:")
            print(repr(g))
            print("expected:")
            print(str(rexp))
            print("actual:")
            print(str(ract))
            print("signatures:")
            print((expected**2 * dx).signature())
            print((actual**2 * dx).signature())
            print()
        assert ract == rexp


def test_apply_single_function_pullbacks_triangle3d():
    cell = Cell("triangle")
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))

    UL2 = FiniteElement("Discontinuous Lagrange", cell, 1, (), l2_piola, L2)
    U0 = FiniteElement("Discontinuous Lagrange", cell, 0, (), identity_pullback, L2)
    U = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V = FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1)
    Vd = FiniteElement("Raviart-Thomas", cell, 1, (2,), contravariant_piola, HDiv)
    Vc = FiniteElement("N1curl", cell, 1, (2,), covariant_piola, HCurl)
    T = FiniteElement("Lagrange", cell, 1, (3, 3), identity_pullback, H1)
    S = SymmetricElement(
        {
            (0, 0): 0,
            (1, 0): 1,
            (2, 0): 2,
            (0, 1): 1,
            (1, 1): 3,
            (2, 1): 4,
            (0, 2): 2,
            (1, 2): 4,
            (2, 2): 5,
        },
        [FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1) for _ in range(6)],
    )
    # (0, 2)-symmetric tensors
    COV2T = FiniteElement("Regge", cell, 0, (2, 2), double_covariant_piola, HEin)
    # (2, 0)-symmetric tensors
    CONTRA2T = FiniteElement("HHJ", cell, 0, (2, 2), double_contravariant_piola, HDivDiv)

    Uml2 = MixedElement([UL2, UL2])
    Um = MixedElement([U, U])
    Vm = MixedElement([U, V])
    Vdm = MixedElement([V, Vd])
    Vcm = MixedElement([Vd, Vc])
    Tm = MixedElement([Vc, T])
    Sm = MixedElement([T, S])

    Vd0 = MixedElement([Vd, U0])  # case from failing ffc demo

    W = MixedElement([S, T, Vc, Vd, V, U])

    ul2 = Coefficient(FunctionSpace(domain, UL2))
    u = Coefficient(FunctionSpace(domain, U))
    v = Coefficient(FunctionSpace(domain, V))
    vd = Coefficient(FunctionSpace(domain, Vd))
    vc = Coefficient(FunctionSpace(domain, Vc))
    t = Coefficient(FunctionSpace(domain, T))
    s = Coefficient(FunctionSpace(domain, S))
    cov2t = Coefficient(FunctionSpace(domain, COV2T))
    contra2t = Coefficient(FunctionSpace(domain, CONTRA2T))

    uml2 = Coefficient(FunctionSpace(domain, Uml2))
    um = Coefficient(FunctionSpace(domain, Um))
    vm = Coefficient(FunctionSpace(domain, Vm))
    vdm = Coefficient(FunctionSpace(domain, Vdm))
    vcm = Coefficient(FunctionSpace(domain, Vcm))
    tm = Coefficient(FunctionSpace(domain, Tm))
    sm = Coefficient(FunctionSpace(domain, Sm))

    vd0m = Coefficient(FunctionSpace(domain, Vd0))  # case from failing ffc demo

    w = Coefficient(FunctionSpace(domain, W))

    rul2 = ReferenceValue(ul2)
    ru = ReferenceValue(u)
    rv = ReferenceValue(v)
    rvd = ReferenceValue(vd)
    rvc = ReferenceValue(vc)
    rt = ReferenceValue(t)
    rs = ReferenceValue(s)
    rcov2t = ReferenceValue(cov2t)
    rcontra2t = ReferenceValue(contra2t)

    ruml2 = ReferenceValue(uml2)
    rum = ReferenceValue(um)
    rvm = ReferenceValue(vm)
    rvdm = ReferenceValue(vdm)
    rvcm = ReferenceValue(vcm)
    rtm = ReferenceValue(tm)
    rsm = ReferenceValue(sm)

    rvd0m = ReferenceValue(vd0m)

    rw = ReferenceValue(w)
    assert len(w) == 9 + 9 + 3 + 3 + 3 + 1
    assert len(rw) == 6 + 9 + 2 + 2 + 3 + 1
    assert len(w) == 28
    assert len(rw) == 23

    assert len(vd0m) == 4
    assert len(rvd0m) == 3

    # Geometric quantities we need:
    J = Jacobian(domain)
    detJ = JacobianDeterminant(domain)
    Jinv = JacobianInverse(domain)
    # o = CellOrientation(domain)
    i, j, k, l = indices(4)  # noqa: E741

    # Contravariant H(div) Piola mapping:
    M_hdiv = (1.0 / detJ) * J  # Not applying cell orientation here
    # Covariant H(curl) Piola mapping: Jinv.T

    mappings = {
        # Simple elements should get a simple representation
        ul2: rul2 / detJ,
        u: ru,
        v: rv,
        vd: as_vector(M_hdiv[i, j] * rvd[j], i),
        vc: as_vector(Jinv[j, i] * rvc[j], i),
        t: rt,
        s: as_tensor([[rs[0], rs[1], rs[2]], [rs[1], rs[3], rs[4]], [rs[2], rs[4], rs[5]]]),
        cov2t: as_tensor(Jinv[k, i] * rcov2t[k, l] * Jinv[l, j], (i, j)),
        contra2t: as_tensor((1.0 / detJ) ** 2 * J[i, k] * rcontra2t[k, l] * J[j, l], (i, j)),
        # Mixed elements become a bit more complicated
        uml2: as_vector([ruml2[0] / detJ, ruml2[1] / detJ]),
        um: rum,
        vm: rvm,
        vdm: as_vector(
            [
                # V
                rvdm[0],
                rvdm[1],
                rvdm[2],
                # Vd
                *(
                    as_tensor(M_hdiv[i, j] * as_vector([rvdm[3], rvdm[4]])[j], (i,))[n]
                    for n in range(3)
                ),
            ]
        ),
        vcm: as_vector(
            [
                # Vd
                *(
                    as_tensor(M_hdiv[i, j] * as_vector([rvcm[0], rvcm[1]])[j], (i,))[n]
                    for n in range(3)
                ),
                # Vc
                *(
                    as_tensor(Jinv[i, j] * as_vector([rvcm[2], rvcm[3]])[i], (j,))[n]
                    for n in range(3)
                ),
            ]
        ),
        tm: as_vector(
            [
                # Vc
                *(
                    as_tensor(Jinv[i, j] * as_vector([rtm[0], rtm[1]])[i], (j,))[n]
                    for n in range(3)
                ),
                # T
                rtm[2],
                rtm[3],
                rtm[4],
                rtm[5],
                rtm[6],
                rtm[7],
                rtm[8],
                rtm[9],
                rtm[10],
            ]
        ),
        sm: as_vector(
            [
                # T
                rsm[0],
                rsm[1],
                rsm[2],
                rsm[3],
                rsm[4],
                rsm[5],
                rsm[6],
                rsm[7],
                rsm[8],
                # S
                rsm[9],
                rsm[10],
                rsm[11],
                rsm[10],
                rsm[12],
                rsm[13],
                rsm[11],
                rsm[13],
                rsm[14],
            ]
        ),
        # Case from failing ffc demo:
        vd0m: as_vector(
            [
                M_hdiv[0, j] * as_vector([rvd0m[0], rvd0m[1]])[j],
                M_hdiv[1, j] * as_vector([rvd0m[0], rvd0m[1]])[j],
                M_hdiv[2, j] * as_vector([rvd0m[0], rvd0m[1]])[j],
                rvd0m[2],
            ]
        ),
        # This combines it all:
        w: as_vector(
            [
                # S
                rw[0],
                rw[1],
                rw[2],
                rw[1],
                rw[3],
                rw[4],
                rw[2],
                rw[4],
                rw[5],
                # T
                rw[6],
                rw[7],
                rw[8],
                rw[9],
                rw[10],
                rw[11],
                rw[12],
                rw[13],
                rw[14],
                # Vc
                *(
                    as_tensor(Jinv[i, j] * as_vector([rw[15], rw[16]])[i], (j,))[n]
                    for n in range(3)
                ),
                # Vd
                *(
                    as_tensor(M_hdiv[i, j] * as_vector([rw[17], rw[18]])[j], (i,))[n]
                    for n in range(3)
                ),
                # V
                rw[19],
                rw[20],
                rw[21],
                # U
                rw[22],
            ]
        ),
    }

    # Check functions of various elements outside a mixed context
    check_single_function_pullback(ul2, mappings)
    check_single_function_pullback(u, mappings)
    check_single_function_pullback(v, mappings)
    check_single_function_pullback(vd, mappings)
    check_single_function_pullback(vc, mappings)
    check_single_function_pullback(t, mappings)
    check_single_function_pullback(s, mappings)
    check_single_function_pullback(cov2t, mappings)
    check_single_function_pullback(contra2t, mappings)

    # Check functions of various elements inside a mixed context
    check_single_function_pullback(uml2, mappings)
    check_single_function_pullback(um, mappings)
    check_single_function_pullback(vm, mappings)
    check_single_function_pullback(vdm, mappings)
    check_single_function_pullback(vcm, mappings)
    check_single_function_pullback(tm, mappings)
    check_single_function_pullback(sm, mappings)

    # Check the ridiculous mixed element W combining it all
    check_single_function_pullback(w, mappings)


def test_apply_single_function_pullbacks_triangle():
    cell = triangle
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))

    Ul2 = FiniteElement("Discontinuous Lagrange", cell, 1, (), l2_piola, L2)
    U = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
    Vd = FiniteElement("Raviart-Thomas", cell, 1, (2,), contravariant_piola, HDiv)
    Vc = FiniteElement("N1curl", cell, 1, (2,), covariant_piola, HCurl)
    T = FiniteElement("Lagrange", cell, 1, (2, 2), identity_pullback, H1)
    S = SymmetricElement(
        {(0, 0): 0, (0, 1): 1, (1, 0): 1, (1, 1): 2},
        [FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1) for i in range(3)],
    )

    Uml2 = MixedElement([Ul2, Ul2])
    Um = MixedElement([U, U])
    Vm = MixedElement([U, V])
    Vdm = MixedElement([V, Vd])
    Vcm = MixedElement([Vd, Vc])
    Tm = MixedElement([Vc, T])
    Sm = MixedElement([T, S])

    W = MixedElement([S, T, Vc, Vd, V, U])

    ul2 = Coefficient(FunctionSpace(domain, Ul2))
    u = Coefficient(FunctionSpace(domain, U))
    v = Coefficient(FunctionSpace(domain, V))
    vd = Coefficient(FunctionSpace(domain, Vd))
    vc = Coefficient(FunctionSpace(domain, Vc))
    t = Coefficient(FunctionSpace(domain, T))
    s = Coefficient(FunctionSpace(domain, S))

    uml2 = Coefficient(FunctionSpace(domain, Uml2))
    um = Coefficient(FunctionSpace(domain, Um))
    vm = Coefficient(FunctionSpace(domain, Vm))
    vdm = Coefficient(FunctionSpace(domain, Vdm))
    vcm = Coefficient(FunctionSpace(domain, Vcm))
    tm = Coefficient(FunctionSpace(domain, Tm))
    sm = Coefficient(FunctionSpace(domain, Sm))

    w = Coefficient(FunctionSpace(domain, W))

    rul2 = ReferenceValue(ul2)
    ru = ReferenceValue(u)
    rv = ReferenceValue(v)
    rvd = ReferenceValue(vd)
    rvc = ReferenceValue(vc)
    rt = ReferenceValue(t)
    rs = ReferenceValue(s)

    ruml2 = ReferenceValue(uml2)
    rum = ReferenceValue(um)
    rvm = ReferenceValue(vm)
    rvdm = ReferenceValue(vdm)
    rvcm = ReferenceValue(vcm)
    rtm = ReferenceValue(tm)
    rsm = ReferenceValue(sm)

    rw = ReferenceValue(w)

    assert len(w) == 4 + 4 + 2 + 2 + 2 + 1
    assert len(rw) == 3 + 4 + 2 + 2 + 2 + 1
    assert len(w) == 15
    assert len(rw) == 14

    # Geometric quantities we need:
    J = Jacobian(domain)
    detJ = JacobianDeterminant(domain)
    Jinv = JacobianInverse(domain)
    i, j, k, l = indices(4)  # noqa: E741

    # Contravariant H(div) Piola mapping:
    M_hdiv = (1.0 / detJ) * J
    # Covariant H(curl) Piola mapping: Jinv.T

    mappings = {
        # Simple elements should get a simple representation
        ul2: rul2 / detJ,
        u: ru,
        v: rv,
        vd: as_vector(M_hdiv[i, j] * rvd[j], i),
        vc: as_vector(Jinv[j, i] * rvc[j], i),
        t: rt,
        s: as_tensor([[rs[0], rs[1]], [rs[1], rs[2]]]),
        # Mixed elements become a bit more complicated
        uml2: as_vector([ruml2[0] / detJ, ruml2[1] / detJ]),
        um: rum,
        vm: rvm,
        vdm: as_vector(
            [
                # V
                rvdm[0],
                rvdm[1],
                # Vd
                *(
                    as_tensor(M_hdiv[i, j] * as_vector([rvdm[2], rvdm[3]])[j], (i,))[n]
                    for n in range(2)
                ),
            ]
        ),
        vcm: as_vector(
            [
                # Vd
                *(
                    as_tensor(M_hdiv[i, j] * as_vector([rvcm[0], rvcm[1]])[j], (i,))[n]
                    for n in range(2)
                ),
                # Vc
                *(
                    as_tensor(Jinv[i, j] * as_vector([rvcm[2], rvcm[3]])[i], (j,))[n]
                    for n in range(2)
                ),
            ]
        ),
        tm: as_vector(
            [
                # Vc
                *(
                    as_tensor(Jinv[i, j] * as_vector([rtm[0], rtm[1]])[i], (j,))[n]
                    for n in range(2)
                ),
                # T
                rtm[2],
                rtm[3],
                rtm[4],
                rtm[5],
            ]
        ),
        sm: as_vector(
            [
                # T
                rsm[0],
                rsm[1],
                rsm[2],
                rsm[3],
                # S
                rsm[4],
                rsm[5],
                rsm[5],
                rsm[6],
            ]
        ),
        # This combines it all:
        w: as_vector(
            [
                # S
                rw[0],
                rw[1],
                rw[1],
                rw[2],
                # T
                rw[3],
                rw[4],
                rw[5],
                rw[6],
                # Vc
                *(as_tensor(Jinv[i, j] * as_vector([rw[7], rw[8]])[i], (j,))[n] for n in range(2)),
                # Vd
                *(
                    as_tensor(M_hdiv[i, j] * as_vector([rw[9], rw[10]])[j], (i,))[n]
                    for n in range(2)
                ),
                # V
                rw[11],
                rw[12],
                # U
                rw[13],
            ]
        ),
    }

    # Check functions of various elements outside a mixed context
    check_single_function_pullback(ul2, mappings)
    check_single_function_pullback(u, mappings)
    check_single_function_pullback(v, mappings)
    check_single_function_pullback(vd, mappings)
    check_single_function_pullback(vc, mappings)
    check_single_function_pullback(t, mappings)
    check_single_function_pullback(s, mappings)

    # Check functions of various elements inside a mixed context
    check_single_function_pullback(uml2, mappings)
    check_single_function_pullback(um, mappings)
    check_single_function_pullback(vm, mappings)
    check_single_function_pullback(vdm, mappings)
    check_single_function_pullback(vcm, mappings)
    check_single_function_pullback(tm, mappings)
    check_single_function_pullback(sm, mappings)

    # Check the ridiculous mixed element W combining it all
    check_single_function_pullback(w, mappings)



"""Algorithms related to restrictions."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction


class RestrictionChecker(MultiFunction):
    """Restiction checker."""

    def __init__(self, require_restriction):
        """Initialise."""
        MultiFunction.__init__(self)
        self.current_restriction = None
        self.require_restriction = require_restriction

    def expr(self, o):
        """Apply to expr."""
        pass

    def restricted(self, o):
        """Apply to restricted."""
        if self.current_restriction is not None:
            raise ValueError("Not expecting twice restricted expression.")
        self.current_restriction = o._side
        (e,) = o.ufl_operands
        self.visit(e)
        self.current_restriction = None

    def facet_normal(self, o):
        """Apply to facet_normal."""
        if self.require_restriction:
            if self.current_restriction is None:
                raise ValueError("Facet normal must be restricted in interior facet integrals.")
        else:
            if self.current_restriction is not None:
                raise ValueError("Restrictions are only allowed for interior facet integrals.")

    def form_argument(self, o):
        """Apply to form_argument."""
        if self.require_restriction:
            if self.current_restriction is None:
                raise ValueError("Form argument must be restricted in interior facet integrals.")
        else:
            if self.current_restriction is not None:
                raise ValueError("Restrictions are only allowed for interior facet integrals.")


def check_restrictions(expression, require_restriction):
    """Check that types that must be restricted are restricted in expression."""
    rules = RestrictionChecker(require_restriction)
    return map_expr_dag(rules, expression)



"""The Measure class."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg 2008-2016
# Modified by Massimiliano Leoni, 2016.

import numbers
from itertools import chain

from ufl.checks import is_true_ufl_scalar
from ufl.constantvalue import as_ufl
from ufl.core.expr import Expr
from ufl.domain import AbstractDomain, as_domain, extract_domains
from ufl.protocols import id_or_none

# Export list for ufl.classes
__all_classes__ = ["Measure", "MeasureSum", "MeasureProduct"]


# TODO: Design a class IntegralType(name, shortname, codim, num_cells, ...)?
# TODO: Improve descriptions below:

# Enumeration of valid domain types
_integral_types = [
    # === Integration over full topological dimension:
    ("cell", "dx"),  # Over cells of a mesh
    # === Integration over topological dimension - 1:
    ("exterior_facet", "ds"),  # Over one-sided exterior facets of a mesh
    ("interior_facet", "dS"),  # Over two-sided facets between pairs of adjacent cells of a mesh
    # === Integration over topological dimension 0
    ("vertex", "dP"),  # Over vertices of a mesh
    # === Integration over custom domains
    ("custom", "dc"),  # Over custom user-defined domains (run-time quadrature points)
    ("cutcell", "dC"),  # Over a cell with some part cut away (run-time quadrature points)
    (
        "interface",
        "dI",
    ),  # Over a facet fragment overlapping with two or more cells (run-time quadrature points)
    (
        "overlap",
        "dO",
    ),  # Over a cell fragment overlapping with two or more cells (run-time quadrature points)
    # === Firedrake specifics:
    ("exterior_facet_bottom", "ds_b"),  # Over bottom facets on extruded mesh
    ("exterior_facet_top", "ds_t"),  # Over top facets on extruded mesh
    ("exterior_facet_vert", "ds_v"),  # Over side facets of an extruded mesh
    ("interior_facet_horiz", "dS_h"),  # Over horizontal facets of an extruded mesh
    ("interior_facet_vert", "dS_v"),  # Over vertical facets of an extruded mesh
]

integral_type_to_measure_name = {i: s for i, s in _integral_types}
measure_name_to_integral_type = {s: i for i, s in _integral_types}

custom_integral_types = ("custom", "cutcell", "interface", "overlap")
point_integral_types = ("vertex",)  # "point")
facet_integral_types = ("exterior_facet", "interior_facet")


def register_integral_type(integral_type, measure_name):
    """Register an integral type."""
    global integral_type_to_measure_name, measure_name_to_integral_type
    if measure_name != integral_type_to_measure_name.get(integral_type, measure_name):
        raise ValueError("Integral type already added with different measure name!")
    if integral_type != measure_name_to_integral_type.get(measure_name, integral_type):
        raise ValueError("Measure name already used for another domain type!")
    integral_type_to_measure_name[integral_type] = measure_name
    measure_name_to_integral_type[measure_name] = integral_type


def as_integral_type(integral_type):
    """Map short name to long name and require a valid one."""
    integral_type = integral_type.replace(" ", "_")
    integral_type = measure_name_to_integral_type.get(integral_type, integral_type)
    if integral_type not in integral_type_to_measure_name:
        raise ValueError("Invalid integral_type.")
    return integral_type


def integral_types():
    """Return a tuple of all domain type strings."""
    return tuple(sorted(integral_type_to_measure_name.keys()))


def measure_names():
    """Return a tuple of all measure name strings."""
    return tuple(sorted(measure_name_to_integral_type.keys()))


class Measure(object):
    """Representation of an integration measure.

    The Measure object holds information about integration properties
    to be transferred to a Form on multiplication with a scalar
    expression.
    """

    __slots__ = ("_integral_type", "_domain", "_subdomain_id", "_metadata", "_subdomain_data")

    def __init__(
        self,
        integral_type,  # "dx" etc
        domain=None,
        subdomain_id="everywhere",
        metadata=None,
        subdomain_data=None,
    ):
        """Initialise.

        Args:
            integral_type: one of "cell", etc, or short form "dx", etc
            domain: an AbstractDomain object (most often a Mesh)
            subdomain_id: either string "everywhere", a single subdomain id int, or tuple of ints
            metadata: dict, with additional compiler-specific parameters
                affecting how code is generated, including parameters
                for optimization or debugging of generated code
            subdomain_data: object representing data to interpret subdomain_id with
        """
        # Map short name to long name and require a valid one
        self._integral_type = as_integral_type(integral_type)

        # Check that we either have a proper AbstractDomain or none
        if domain is not None:
            domain = as_domain(domain)
            if not isinstance(domain, AbstractDomain):
                raise ValueError("Invalid domain.")
        self._domain = domain

        # Store subdomain data
        self._subdomain_data = subdomain_data
        # FIXME: Cannot require this (yet) because we currently have
        # no way to implement ufl_id for dolfin SubDomain
        # if not (self._subdomain_data is None or hasattr(self._subdomain_data, "ufl_id")):
        #     raise ValueError("Invalid domain data, missing ufl_id() implementation.")

        # Accept "everywhere", single subdomain, or multiple
        # subdomains
        if isinstance(subdomain_id, tuple):
            for did in subdomain_id:
                if not isinstance(did, numbers.Integral):
                    raise ValueError(f"Invalid subdomain_id {did}.")
        else:
            if not (subdomain_id in ("everywhere",) or isinstance(subdomain_id, numbers.Integral)):
                raise ValueError(f"Invalid subdomain_id {subdomain_id}.")
        self._subdomain_id = subdomain_id

        # Validate compiler options are None or dict
        if metadata is not None and not isinstance(metadata, dict):
            raise ValueError("Invalid metadata.")
        self._metadata = metadata or {}

    def integral_type(self):
        """Return the domain type.

        Valid domain types are "cell", "exterior_facet", "interior_facet", etc.
        """
        return self._integral_type

    def ufl_domain(self):
        """Return the domain associated with this measure.

        This may be None or a Domain object.
        """
        return self._domain

    def subdomain_id(self):
        """Return the domain id of this measure (integer)."""
        return self._subdomain_id

    def metadata(self):
        """Return the integral metadata.

        This data is not interpreted by UFL.
        It is passed to the form compiler which can ignore it or use
        it to compile each integral of a form in a different way.
        """
        return self._metadata

    def reconstruct(
        self, integral_type=None, subdomain_id=None, domain=None, metadata=None, subdomain_data=None
    ):
        """Construct a new Measure object with some properties replaced with new values.

        Example:
            <dm = Measure instance>
            b = dm.reconstruct(subdomain_id=2)
            c = dm.reconstruct(metadata={ "quadrature_degree": 3 })

        Used by the call operator, so this is equivalent:
            b = dm(2)
            c = dm(0, { "quadrature_degree": 3 })
        """
        if subdomain_id is None:
            subdomain_id = self.subdomain_id()
        if domain is None:
            domain = self.ufl_domain()
        if metadata is None:
            metadata = self.metadata()
        if subdomain_data is None:
            subdomain_data = self.subdomain_data()
        return Measure(
            self.integral_type(),
            domain=domain,
            subdomain_id=subdomain_id,
            metadata=metadata,
            subdomain_data=subdomain_data,
        )

    def subdomain_data(self):
        """Return the integral subdomain_data.

        This data is not interpreted by
        UFL.  Its intension is to give a context in which the domain
        id is interpreted.
        """
        return self._subdomain_data

    # Note: Must keep the order of the first two arguments here
    # (subdomain_id, metadata) for backwards compatibility, because
    # some tutorials write e.g. dx(0, {...}) to set metadata.
    def __call__(
        self,
        subdomain_id=None,
        metadata=None,
        domain=None,
        subdomain_data=None,
        degree=None,
        scheme=None,
    ):
        """Reconfigure measure with new domain specification or metadata."""
        # Let syntax dx() mean integral over everywhere
        all_args = (subdomain_id, metadata, domain, subdomain_data, degree, scheme)
        if all(arg is None for arg in all_args):
            return self.reconstruct(subdomain_id="everywhere")

        # Let syntax dx(domain) or dx(domain, metadata) mean integral
        # over entire domain.  To do this we need to hijack the first
        # argument:
        if subdomain_id is not None and (
            isinstance(subdomain_id, AbstractDomain) or hasattr(subdomain_id, "ufl_domain")
        ):
            if domain is not None:
                raise ValueError(
                    "Ambiguous: setting domain both as keyword argument and first argument."
                )
            subdomain_id, domain = "everywhere", subdomain_id

        # If degree or scheme is set, inject into metadata. This is a
        # quick fix to enable the dx(..., degree=3) notation.
        # TODO: Make degree and scheme properties of integrals instead of adding to metadata.
        if (degree, scheme) != (None, None):
            metadata = {} if metadata is None else metadata.copy()
            if degree is not None:
                metadata["quadrature_degree"] = degree
            if scheme is not None:
                metadata["quadrature_rule"] = scheme

        # If we get any keywords, use them to reconstruct Measure.
        # Note that if only one argument is given, it is the
        # subdomain_id, e.g. dx(3) == dx(subdomain_id=3)
        return self.reconstruct(
            subdomain_id=subdomain_id,
            domain=domain,
            metadata=metadata,
            subdomain_data=subdomain_data,
        )

    def __str__(self):
        """Format as a string."""
        name = integral_type_to_measure_name[self._integral_type]
        args = []

        if self._subdomain_id is not None:
            args.append("subdomain_id=%s" % (self._subdomain_id,))
        if self._domain is not None:
            args.append("domain=%s" % (self._domain,))
        if self._metadata:  # Stored as {} if None
            args.append("metadata=%s" % (self._metadata,))
        if self._subdomain_data is not None:
            args.append("subdomain_data=%s" % (self._subdomain_data,))

        return "%s(%s)" % (name, ", ".join(args))

    def __repr__(self):
        """Return a repr string for this Measure."""
        args = []
        args.append(repr(self._integral_type))

        if self._subdomain_id is not None:
            args.append("subdomain_id=%s" % repr(self._subdomain_id))
        if self._domain is not None:
            args.append("domain=%s" % repr(self._domain))
        if self._metadata:  # Stored as {} if None
            args.append("metadata=%s" % repr(self._metadata))
        if self._subdomain_data is not None:
            args.append("subdomain_data=%s" % repr(self._subdomain_data))

        r = "%s(%s)" % (type(self).__name__, ", ".join(args))
        return r

    def __hash__(self):
        """Return a hash value for this Measure."""
        metadata_hashdata = tuple(sorted((k, id(v)) for k, v in list(self._metadata.items())))
        hashdata = (
            self._integral_type,
            self._subdomain_id,
            hash(self._domain),
            metadata_hashdata,
            id_or_none(self._subdomain_data),
        )
        return hash(hashdata)

    def __eq__(self, other):
        """Checks if two Measures are equal."""
        sorted_metadata = sorted((k, id(v)) for k, v in list(self._metadata.items()))
        sorted_other_metadata = sorted((k, id(v)) for k, v in list(other._metadata.items()))

        return (
            isinstance(other, Measure)
            and self._integral_type == other._integral_type
            and self._subdomain_id == other._subdomain_id
            and self._domain == other._domain
            and id_or_none(self._subdomain_data) == id_or_none(other._subdomain_data)
            and sorted_metadata == sorted_other_metadata
        )

    def __add__(self, other):
        """Add two measures (self+other).

        Creates an intermediate object used for the notation
          expr * (dx(1) + dx(2)) := expr * dx(1) + expr * dx(2)
        """
        if isinstance(other, Measure):
            # Let dx(1) + dx(2) equal dx((1,2))
            return MeasureSum(self, other)
        else:
            # Can only add Measures
            return NotImplemented

    def __mul__(self, other):
        """Multiply two measures (self*other).

        Creates an intermediate object used for the notation
          expr * (dm1 * dm2) := expr * dm1 * dm2

        This is work in progress and not functional.
        """
        if isinstance(other, Measure):
            # Tensor product measure support
            return MeasureProduct(self, other)
        else:
            # Can't multiply Measure from the right with non-Measure type
            return NotImplemented

    def __rmul__(self, integrand):
        """Multiply a scalar expression with measure to construct a form with a single integral.

        This is to implement the notation
            form = integrand * self

        Integration properties are taken from this Measure object.
        """
        # Avoid circular imports
        from ufl.form import Form
        from ufl.integral import Integral

        # Allow python literals: 1*dx and 1.0*dx
        if isinstance(integrand, (int, float)):
            integrand = as_ufl(integrand)

        # Let other types implement multiplication with Measure if
        # they want to (to support the dolfin-adjoint TimeMeasure)
        if not isinstance(integrand, Expr):
            return NotImplemented

        # Allow only scalar integrands
        if not is_true_ufl_scalar(integrand):
            raise ValueError(
                "Can only integrate scalar expressions. The integrand is a "
                f"tensor expression with value shape {integrand.ufl_shape} and "
                f"free indices with labels {integrand.ufl_free_indices}."
            )

        # If we have a tuple of domain ids build the integrals one by
        # one and construct as a Form in one go.
        subdomain_id = self.subdomain_id()
        if isinstance(subdomain_id, tuple):
            return Form(
                list(
                    chain(
                        *(
                            (integrand * self.reconstruct(subdomain_id=d)).integrals()
                            for d in subdomain_id
                        )
                    )
                )
            )

        # Check that we have an integer subdomain or a string
        # ("everywhere" or "otherwise", any more?)
        if not isinstance(
            subdomain_id,
            (
                str,
                numbers.Integral,
            ),
        ):
            raise ValueError("Expecting integer or string domain id.")

        # If we don't have an integration domain, try to find one in
        # integrand
        domain = self.ufl_domain()
        if domain is None:
            domains = extract_domains(integrand)
            if len(domains) == 1:
                (domain,) = domains
            elif len(domains) == 0:
                raise ValueError("This integral is missing an integration domain.")
            else:
                raise ValueError(
                    "Multiple domains found, making the choice of integration domain ambiguous."
                )

        # Otherwise create and return a one-integral form
        integral = Integral(
            integrand=integrand,
            integral_type=self.integral_type(),
            domain=domain,
            subdomain_id=subdomain_id,
            metadata=self.metadata(),
            subdomain_data=self.subdomain_data(),
        )
        return Form([integral])


class MeasureSum(object):
    """Represents a sum of measures.

    This is a notational intermediate object to translate the notation
        f*(ds(1)+ds(3))
    into
        f*ds(1) + f*ds(3)
    """

    __slots__ = ("_measures",)

    def __init__(self, *measures):
        """Initialise."""
        self._measures = measures

    def __rmul__(self, other):
        """Multiply."""
        integrals = [other * m for m in self._measures]
        return sum(integrals)

    def __add__(self, other):
        """Add."""
        if isinstance(other, Measure):
            return MeasureSum(*(self._measures + (other,)))
        elif isinstance(other, MeasureSum):
            return MeasureSum(*(self._measures + other._measures))
        return NotImplemented

    def __str__(self):
        """Format as a string."""
        return "{\n    " + "\n  + ".join(map(str, self._measures)) + "\n}"


class MeasureProduct(object):
    """Represents a product of measures.

    This is a notational intermediate object to handle the notation

        f*(dm1*dm2)

    This is work in progress and not functional. It needs support
    in other parts of ufl and the rest of the code generation chain.
    """

    __slots__ = ("_measures",)

    def __init__(self, *measures):
        """Create MeasureProduct from given list of measures."""
        self._measures = measures
        if len(self._measures) < 2:
            raise ValueError("Expecting at least two measures.")

    def __mul__(self, other):
        """Flatten multiplication of product measures.

        This is to ensure that (dm1*dm2)*dm3 is stored as a simple
        list (dm1,dm2,dm3) in a single MeasureProduct.
        """
        if isinstance(other, Measure):
            measures = self.sub_measures() + [other]
            return MeasureProduct(*measures)
        else:
            return NotImplemented

    def __rmul__(self, integrand):
        """Multiply."""
        # TODO: Implement MeasureProduct.__rmul__ to construct integral and form somehow.
        raise NotImplementedError()

    def sub_measures(self):
        """Return submeasures."""
        return self._measures
